//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 18:18:08 sutner>
// 
// File:     bitvector.cc
// Purpose:  BitVector's and basic linear algebra over GF(2)
//           use arrays of size_ts to store bit vectors
//           also 2-dim matrices
//--------------------------------------------------------------------------------

#include "basic.h"
#include "bitvect.h"

// replace by algo.h
#define    MAX(x,y)    ((x)>(y))?(x):(y)
#define    DIV(x,y)    (((x)%(y)==0)? (x)/(y) : ((x)/(y))+1 )
#define    MOD_(x,y)   ((x)%(y)==0)?(y):((x)%(y))

//----------------------------------------------------------------------
//  BitVector
//  packed bit-vectors
//----------------------------------------------------------------------
// Initialize/Finalize 

BitVector::BitVector(): numbits(0), numblocks(0), uselast(0), blocks(NULL) {}

BitVector::BitVector( int sz ):
    numbits(sz), numblocks(DIV(sz,BLOCKSIZE)), 
    uselast(MOD_(sz, BLOCKSIZE)), blocks(new size_t[numblocks]) {}


BitVector::BitVector( int sz, size_t initblock ):
    numbits(sz), numblocks(DIV(sz,BLOCKSIZE)), 
    uselast(MOD_(sz, BLOCKSIZE)), blocks(new size_t[numblocks])
{
   for (int i = 0; i < numblocks; i++) 
     blocks[i] = initblock;
   // put padding 0's in last block
   blocks[numblocks - 1] &= (UINT_MAX >> (BLOCKSIZE - uselast));
}

BitVector::~BitVector() { delete [] blocks; }

void  BitVector::destroy()
{
   delete [] blocks;
   blocks = NULL;
   numbits = numblocks = uselast = 0;
}

void  BitVector::initialize( int sz )
{
   delete [] blocks;
   numbits   = sz; 
   numblocks = DIV(sz,BLOCKSIZE); 
   uselast   = MOD_(sz, BLOCKSIZE);
   blocks    = new size_t[numblocks];
}

				// only for 0/1 initialization
void  BitVector::initialize( int sz, int b )
{
   initialize(sz); 
   int i;
   
   if(b)
     for (i = 0; i < numblocks; i++) 
       blocks[i] = ONE_BLOCK;
   else
     for (i = 0; i < numblocks; i++) 
       blocks[i] = 0U;
}


void  BitVector::initialize( int sz, VEC_FN initfn )
{
   initialize( sz );
   
   // initialize bits using initfn:
   for (int i = 0; i < numbits; i++) 
     setbit(i, (*initfn)(i, numbits));
}


// Copy 

BitVector::BitVector( const BitVector& vec ):
    numbits(vec.numbits), numblocks(vec.numblocks),
    uselast(vec.uselast), blocks(new size_t[numblocks])
{
   memcpy(blocks, vec.blocks, numblocks * sizeof(size_t));
}

BitVector  BitVector::operator=( const BitVector& vec )
{
   if (this == &vec) 
     return *this;

   if (numblocks != vec.numblocks)
   {   // free old elts and reallocate
      delete [] blocks;
      numblocks = vec.numblocks;
      blocks = new size_t[numblocks];
   }

   numbits   = vec.numbits;
   uselast   = vec.uselast;
   
   memcpy(blocks, vec.blocks, numblocks * sizeof(size_t));

   return *this;
}


size_t BitVector::get(int n) const
{
   if (0 <= n && n < numblocks)
     return blocks[n];
   else
     return 0U;
}

// Attributes 

int BitVector::size() const { return numbits; }

int BitVector::numBlocks() const { return numblocks; }

// Input/Output 
ostream& operator<<(ostream& strm, const BitVector& vec)
{
    size_t block, mask, *p = vec.blocks;
    int i, j;

    if( vec.numbits == 0)
	return strm;

    for( i = 0; i < vec.numblocks - 1; i++ ) 
    {
	block = *p++;
	mask = 1U;
	for( j = 0; j < BLOCKSIZE; j++ )
	{
	    if( block & mask )
		strm << ON_CHAR;
	    else
		strm << OFF_CHAR;
	    mask <<= 1;
	}
    }
    block = *p;
    mask = 1U;
    for( j = 0; j < vec.uselast; j++ )
    {
	if( block & mask )
	    strm << ON_CHAR;
	else
	    strm << OFF_CHAR;
	mask <<= 1;
    }
   
    return strm;
}

//----------------------------------------------------------------------
//  algebraic operations
//----------------------------------------------------------------------

// left shift a vector
// cyc == 0: fixed boundary conditions, cyclic otherwise
BitVector  BitVector::left_shift( int cyc ) const
{
    int i;
    BitVector  res(numbits);
    size_t carry0 = 0U, carry1;
    size_t mask = UINT_MAX >> (BLOCKSIZE-uselast); //mask for last block
   
    for( i=0; i<numblocks-1; i++ )
    {
	carry1 = (blocks[i] & HIBIT) ? 1U : 0U;
	res.blocks[i] = (blocks[i]<<1) | carry0;
	carry0 = carry1;
    }
				// i == numblocks-1
    if( cyc==0 )
	res.blocks[i] = ((blocks[i]<<1) | carry0) & mask;
    else
    {
	carry1 = ( blocks[i] & (1U << (uselast-1))) ? 1U : 0U;
	res.blocks[i] = ((blocks[i]<<1) | carry0) & mask;
	res.blocks[0] |= carry1;
    }
    return res;
}

 
// right shift a vector
// cyc == 0: fixed boundary conditions, cyclic otherwise
BitVector  BitVector::right_shift( int cyc ) const
{
    int i;
    BitVector res(numbits);
    size_t carry0 = 0U, carry1;

    for( i=numblocks-1; i>0; i-- )
    {
	carry1 = (blocks[i] & 1U) ? HIBIT : 0U;
	res.blocks[i] = (blocks[i]>>1) | carry0;
	carry0 = carry1;
    }
   
    if( cyc==0 )
	res.blocks[0] = (blocks[0]>>1) | carry0;
    else
    {
	carry1 = (blocks[0] & 1U) ? (1U << (uselast-1)) : 0U;
	res.blocks[0] = (blocks[0]>>1) | carry0;
	res.blocks[numblocks-1] |= carry1;
    }
   
    return res;
}


// add two vectors, possibly of different size
BitVector  BitVector::operator+( const BitVector& b ) const 
{
   BitVector  c(MAX(numbits,b.numbits));
   
   for( int i = 0; i < c.numblocks; i++ )
     c.blocks[i] = get(i) ^ b.get(i); // note: get(i) defaults to 0
   
   return c;
}

// dot product of two vectors, possibly of different size
int  BitVector::operator*( const BitVector& b ) const 
{
   int  end = MAX(numblocks,b.numblocks), res = 0, i;
   size_t _xor = 0U;               // cumulative XOR of AND of corresponding blocks.
   
   for( i = 0; i < end; i++ )
     _xor ^= get(i) & b.get(i);  // note: get(i) defaults to 0
   
   for( i = 0; i < BLOCKSIZE; i++ )
   {
      res ^= _xor & 1U;
      _xor >>= 1;
   }

   return res;
}

BitVector  BitVector::sigma( int cyc ) const
{
   return left_shift(cyc) + right_shift(cyc);
}

BitVector  BitVector::sigma_plus( int cyc ) const 
{
   return left_shift(cyc) + right_shift(cyc) + *this;
}



//---------------------------------------------------------------------------
// BitMatrix
// arrays of BitVectors (pointers thereto)
//---------------------------------------------------------------------------


BitMatrix::BitMatrix( int n ): _rows(n), _numcols(n)
{ 
   for( int i=0; i<n; i++ )
     _rows[i] = new BitVector(n);
}

BitMatrix::BitMatrix( int n, int m ): _rows(n), _numcols(m)
{ 
   for( int i=0; i<n; i++ )
     _rows[i] = new BitVector(m);
}

BitMatrix::BitMatrix( int n, int m, int bb ): _rows(n), _numcols(m)
{ 
   for( int i=0; i<n; i++ )
     _rows[i] = new BitVector(m,bb);
}

BitMatrix::~BitMatrix()
{
   for( int i=0; i<numRows(); i++ )
     delete _rows[i];
}


void BitMatrix::initialize( int rr, int cc, MAT_FN initfn )
{ 
   int i,j;
   _numcols = cc;  
   _rows.resize( rr );

   for( i=0; i<rr; i++ )
     _rows[i] = new BitVector(cc);

   for( i=0; i<rr; i++ )
     for( j=0; j<cc; j++ ) 
       _rows[i]->setbit( j, (*initfn)( i, j, rr, cc ));
}




BitMatrix::BitMatrix( const BitMatrix& src ): 
   _rows(src._rows.size()), _numcols(src._numcols) 
{ 
   for( size_t i=0; i<_rows.size(); i++ )
     *_rows[i] = *src._rows[i];
}


BitMatrix&  BitMatrix::operator=( const BitMatrix& src )
{
   if( this==&src ) 
     return *this;

   if( numRows()!=src.numRows() || numColumns()!=src.numColumns() )
   {  
      _numcols = src.numColumns();
      _rows.resize( src.numRows() );
      for( int i=0; i<numRows(); i++ )
	_rows[i] = new BitVector(_numcols);
   }

   for( int i=0; i<numRows(); i++ )
     *_rows[i] = *src._rows[i];

   return *this;
}


void   BitMatrix::swap_rows( int i, int j )
{
   BitVector  *tmp;
   tmp  = _rows[i];
   _rows[i] = _rows[j];
   _rows[j] = tmp;
}

// add row i to j 
void   BitMatrix::add_rows( int i, int j )
{
   *_rows[j] = *_rows[i] + *_rows[j];	// bad, uses copy
}


// compute the rank of a m by m matrix  
int  BitMatrix::matrixRank()
{
   int  j, k, i; 
   int  m = numRows();
   int  r = numRows();

   for ( j = 0; j < m; j++ )
   {
      i = j;
      while ( (i < m) && !_rows[i]->getbit(j)  ) 
        i++;
      if (i == m)
        r--;            //  no 1 in j-th column  
      else
      {
         if (i > j) swap_rows( i, j );
         for  ( k = 0; k < j; k++ )
           if ( _rows[k]->getbit(j) ) add_rows( j, k );
         for  ( k = j+1; k < m; k++ )
           if ( _rows[k]->getbit(j) ) add_rows( j, k );
      }
   }
   
   return  r;
}

ostream& operator<<( ostream& strm, BitMatrix& bm )
{
   for( int i=0; i<bm.numRows(); i++ )
     strm << *bm[i] << endl;
   
   return strm << endl;
}



//---------------------------------------------------------------------------
