//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 18:20:03 sutner>
// 
// File:     fsmdfa.cc
// Purpose:  deterministic complete FAs
//
//--------------------------------------------------------------------------------

#include  <stack>

#include  "stl-wrap.h"
#include  "fsm.h"
#include  "urlist.h"
#include  "listutil.h"
#include  "util.h"
#include  "semimod.h"

//----------------------------------------------------------------------
//  equality testing
//  
// note: property flags are ignored in equality testing 
//----------------------------------------------------------------------

DFA::DFA( int n, int k ): FSMachine( n, k )
{
    if(_n) 
    {
	trmat.resize( _n * _k );
	finst.initialize( _n );
	assert( trmat.capacity() == _n * _k );
	assert( finst.capacity() == _n );
    }
}

DFA::DFA( const DFA& M ):
    FSMachine(M._n,M._k,M._pr,M._st), 
    trmat(M.trmat), inist(M.inist), finst(M.finst)  {}


DFA::~DFA() {}


bool  DFA::operator==( const DFA& src ) const 
{
   if( &src==this )		// machines identical 
     return 1;

   if( numStates()  != src.numStates()  || 
       numSymbols() != src.numSymbols() || 
       numFinal()   != src.numFinal()   || 
       initial()    != src.initial() )
     return 0;
   
   if( !(transitions()==src.transitions()) )
      return 0;

   if( !(final()==src.final()) )
      return 0;

   return 1;
}



//----------------------------------------------------------------------
//  accessible parts 


size_t  DFA::accessiblePart( DFA& odfa, bool norm ) const 
{
   size_t  i, s, m;
   int     cnt;
   size_t n = numStates();

   Graph gr(n);
   toGraph( gr );

   IntSet        reach;
   m = gr.weakComponent( initial(), reach );
   
				// renumber reachable states  h : R <--> [m] 
				// h partial bijection, hi inverse
   ivector  h(n), hi(m);
   cnt = 0;   
   IntSet::iterator  it;
   foreach_item( it, reach )
   {		
      h[*it]    = cnt;
      hi[cnt++] = *it;
   }

				// initialize output machine 
   odfa.setdfa( m, _k );
   odfa.accessible() = TRUE;
   odfa.transitions().resize( m*_k );
   
   for( s=0; s<_k; s++ )
     for( i=0; i<m; i++ )
       odfa.delta(i,s) = h[ delta( hi[i], s ) ];


   odfa.initial() = 0;
   odfa.final().initialize(m);
   for( i=0; i<m; i++ )
     if( final().find( hi[i] ) )
	odfa.final().insert(i);

   if( !norm )
     odfa.states() = order_select_ulist( stateSet(), reach, n );

   return m;
}

   
//--------------------------------------------------------------------------------
// conversions

void   DFA::toGraph( Graph& gr ) const 
{
    size_t   p, s, q;

    gr.initialize(_n,0,0);

    foreach_symbol( s, *this )
	foreach_state( p, *this )
	{
	    q = delta(p,s);
	    if( !gr.edgeQ(p,q) )
		gr.insertEdge(p,q);
	}
}
   
void    DFA::toSA( SA& sa, bool norm ) const
{
    sa.set( numStates(), numSymbols() );
    sa.transitions().resize(0);
    
    for( size_t p = 0; p < numStates(); ++p )
	for( size_t s = 0; s < numSymbols(); ++s )
	    sa.transitions().push_back( Trans( p, s, delta(p,s) ) );

    if( !norm )
	sa.states() = states();
}
			    
	
//--------------------------------------------------------------------------------
// accessible part of full product automaton

struct  prod_trans: public binary_function<uupair,size_t,uupair>
{
    const uvector  &T1, &T2;
    size_t   kk, nn1, nn2;
    
    prod_trans( const uvector& t1, const uvector& t2, size_t k, size_t n1, size_t n2 ):
	T1(t1), T2(t2), kk(k), nn1(n1), nn2(n2) {}

    uupair  operator()( uupair p, size_t s ) const 
    {
	return  uupair( T1[ s*nn1 + p.first ], T2[ s*nn2 + p.second ] );
    }
};

// pre:   C is a container holding uupair's 
// note:  states sets are 1-indexed

template<class Cont>
UList   cp_states( size_t n1, size_t n2, UList A, UList B, const Cont& C )
{
    UList  res, x;
    vector<UList>  v1(n1), v2(n2), v(C.size());
    v1.resize(0);
    v2.resize(0);
    while( A )
    {
	A >> x; v1.push_back(x);
    }
    while( B )
    {
	B >> x; v2.push_back(x);
    }

    typename  Cont::const_iterator  it;
    for( it = C.begin(); it != C.end(); ++it )
	v[it->second] = ( v1[ it->first.first ] -= v2[ it->first.second ] -= nil );

    for( int i = v.size()-1; i >= 0; --i )
	res += v[i];

    return  res; 
}


size_t  DFA::productAutomaton( DFA& M, DFA& mout, bool un, bool norm ) const
{
    size_t  k  = numSymbols();
    size_t  n1 = numStates();
    size_t  n2 = M.numStates();

    prod_trans   pt( transitions(), M.transitions(), k, n1, n2 );

    SemiModule<uupair,prod_trans> SM( k, uupair(0,0), pt );

    size_t  n  = SM.generate();

    mout.setdfa( n, k );
				// transpose trace list to transition vector
    ulist::iterator  lit = SM.del.begin();
    for( size_t p = 0; p < n; ++p )
	for( size_t s = 0; s < k; ++s )
	    mout.transitions()[ n*s + p ] = *lit++;

    mout.initial() = SM.Q[ uupair( initial(), M.initial() ) ];
    IntSet::const_iterator  i, j;
    SemiModule<uupair,prod_trans>::iterator it;

    if( un )			             // F1 x F2
    {	
	foreach_item( i, final() )
	    foreach_item( j, M.final() )
	    {
		it = SM.Q.find( uupair(*i,*j) );
		if( it != SM.Q.end() )
		    mout.final().insert( SM.Q[ uupair(*i,*j) ] );
	    }
    }
    else				     // F1 x Q2 u Q1 x F2
    {
	foreach_item( i, final() )
	    for( size_t p = 0; p < n2; ++p )
	    {
		uupair  fred( *i, p );
		it = SM.Q.find( fred );
		if( it != SM.Q.end() )
		    mout.final().insert( SM.Q[ uupair(*i, p ) ] );
	    }
	foreach_item( j, M.final() )
	    for( size_t p = 0; p < n1; ++p )
	    {
		it = SM.Q.find( uupair( p, *j ) );
		if( it != SM.Q.end() )
		    mout.final().insert( SM.Q[ uupair( p, *j ) ] );
	    }
    }

    mout.final().sort();
	
    if( !norm )
	mout.states() = cp_states( n1, n2, stateSet(), M.stateSet(), SM.Q );

    return n;
}


// states() are 1-indexed, have to shift for internal representation

size_t  DFA::quotientAutomaton( DFA& M, FA& fa, bool norm ) const
{
    size_t  k  = numSymbols();
    size_t  n2 = M.numStates();

    DFA  mprod;
    productAutomaton( M, mprod, 0, 0 );
    
    fa.set( n2, k );
    
    UList  Q = mprod.states(), x;
    
    while( Q )
    {
	Q >> x;
	if( final().find( ivalue( x.head() ) - 1 ) )
	    fa.initial().insert( ivalue( x.tail().head() ) - 1 );
    }
    
    fa.final() = M.final();

    size_t p, s;
    foreach_state( p, M )
	foreach_symbol( s, M )
	fa.transitionsInsert( p, s, M.delta(p,s) );
    
    return n2;
}



//----------------------------------------------------------------------
// Hopcroft's minimization algorithm 
//----------------------------------------------------------------------

size_t   DFA::minimalAutomaton( DFA& outdfa, bool norm ) const 
{
    // npart: ultimately the number of states in new machine
    int    j=0, z, zi, npart, oldnpart, ii;
    size_t   i, s;
    
    size_t n  = numStates();
    size_t k  = numSymbols();
    size_t nf = numFinal();
    size_t q0 = initial();
   
    if( nf==0 || nf==n )		// get out if minimal machine is trivial 
    {
	make_basic_dfa( k, (nf)?1:0, outdfa );
	return 1;
    }
   
    ivector        partind(n);
    ivector        candcard(n);
    ivector        inter(n);
    Fill( partind, 0 );
    Fill( candcard, 0 );
    Fill( inter, 0 );

    typedef  Pair<size_t,size_t>  iipair;

    IntSet          iset(n);		        // usage:  iset = f^-1(part[i]) 

    stack<iipair, vector<iipair> >  cand;       // stack of candidates:  pairs (i,s)
    stack<int, ivector >            hits;	// stack of hits
    vector<ulist>                   part(n);	// partition vector 
    vector<ulist>               deltainv(k*n);  // inverse transition function  
   
				// compute inverse of delta function 
    foreach_symbol( s, *this )
	foreach_state( i, *this )
	deltainv[s*n + delta(i,s)].push_front(i);

				// initial partition
    npart = oldnpart = 2;
    for( i=0; i<n; i++ )	
    {
	ii = final().find(i);
	part[ii].push_front(i);
	partind[i] = ii;
    }
   
   
    ii = ( n > 2*nf )? 0 : 1;	// determine which part is handled first
    candcard[0] = candcard[1] = k; // number of candidates k for each part

    for( s=0; s<k; s++ )
	cand.push( iipair(ii,s) );

    for( s=0; s<k; s++ )
	cand.push( iipair(1-ii,s) );

    ulist::iterator  itr, itrr;
   
    while( !cand.empty() )	// main loop ------------------------------- 
    {
	cand.top().get(i,s);		  // split wrto part[i], symbol s 
	cand.pop();
      
	candcard[i]--;
	int ns = n * s;
	foreach_item( itr, part[i] )		  // iset = f^-1(part[i]) 
	    foreach_item( itrr, deltainv[ ns + *itr ] )	
	    {
		iset.insert( *itrr );
		zi = partind[ *itrr ];
		if( inter[zi]==0 ) hits.push( zi );	// parts hit by iset 
		inter[zi]++;
	    }
      
	 
	while( !hits.empty() )		  // for all part[j] touched by iset 
	{
	    j = hits.top();
	    hits.pop();
	    ulist  &pj = part[j];

	    if( size_t(inter[j]) < pj.size() ) // split part[j] by iset 
	    {
		for( itr = pj.begin(); itr != pj.end();  )
		    if( !iset.find( z = *itr ) )
		    {
			itrr = itr;
			++itrr;
			pj.erase(itr);
			itr = itrr;
			part[npart].push_front(z);
			partind[z] = npart;
		    }
		    else
			++itr;

		int nj = ( candcard[j] || pj.size() > part[npart].size() ) ? npart : j;

		candcard[nj] = k;
		for( s = 0; s < k; ++s )
		    cand.push( iipair( nj, s ) );

		npart++;

	    } // end split 
	}	// end touched 
	
	while( !iset.empty() )		  // destroy iset, inter 
	{
	    z  = iset.pop();
	    zi = partind[z];
	    inter[ ( zi >= oldnpart ) ? j : zi ] = 0;
	}
	oldnpart = npart;
      
    } // end main loop ------------------------------------------------


    // construct output machine 
    outdfa.setdfa( npart, k );
   
    IntSet::const_iterator  finitr;
   
    foreach_item( finitr, final() )
	outdfa.final().insert( partind[*finitr] );

    outdfa.final().sort();

    foreach_symbol( s, *this )
	foreach_state( i, *this )
	outdfa.transitions()[s*npart + partind[i]] = partind[delta(i,s)];

    outdfa.initial() = partind[q0];

    if( !norm )
	outdfa.states() =  partition_ulist( stateSet(), partind, n, npart );
   
    outdfa.minimal() = outdfa.accessible() = TRUE; 

    return npart;
}				


  

///---------------------------------------------------------------------
// constructors
				// all/nothing DFA on 1 state
void   make_basic_dfa( int k, int tt, DFA& dfa )
{
   dfa.setdfa( 1, k );
   Fill( dfa.transitions(), 0 );
   dfa.initial() = 0;

   if( tt )
     dfa.final().insert(0);
}


//---------------------------------------------------------------------
// generate acceptance language for DFAs
// unabridged brute force
//---------------------------------------------------------------------

// 				// generate lexicographically next word
// 				// as ivector of size n, base alphabet k  
// static  int  next_word( ivector& A, int k  )
// {
//    int  k1 = k-1;
//    ivector::iterator  ip;

//    ip = A.end() - 1;	  // point at last symbol

//    while( ip!=A.begin() && *ip==k1 )	  // skip back over symbols (k-1) 
//      --ip;

//    if( ip==A.begin() && *ip==k1 )	  // last word (k-1)^n reached, exit 
//      return 0;

//    (*ip++)++;				  // increment symbol 
   
//    while( ip!=A.end() )	          // set rest to 0 
//      *ip++ = 0;

//    return 1;
// }

//--------------------------------------------------------------------
// search for a sink

int  sink_dfa( const DFA& dfa )
{
    size_t n = dfa.numStates();
    size_t k = dfa.numSymbols();
    
    size_t   p, s, sink;
    
    for( p=0; p<n; p++ )
    {
	if( dfa.final().find(p) )
	    continue;
	
	sink = 1;
	for( s=0; s<k && sink; s++ )
	    if( dfa.delta(p,s) != p )
		sink = 0;
	if( sink )
	    return p;
    }
    return -1;
}


//----------------------------------------------------------------------
//  matrix  L:  [0,1] x Q
//  where  L[t,p] = { w |  w : q0 --> p, |w|=t } 

UList   DFA::language( int rr ) const  
{
    UList  res, LL;
    int    t, p, s;
    int    all = 0;
    int    k  = numSymbols();
    int    n  = numStates();
    int    q0 = initial();
    int    sk = sink_dfa(*this);
    

    if( rr<0 )			// negative: up to words of size -rr 
	rr = -rr, all = 1;

   
				// back and forth copying
    vector<list<string> >      L1(n), L2(n);
    vector<list<string> >      *p1, *p2, *tmp;
    p1 = &L1;
    p2 = &L2;

    p1->operator[](q0).push_back("");
    if( all )
    {
	if( final().find(q0) ) 
	    res += new StringAtom(""); 
	LL += res;
    }
    

    list<string>::const_iterator   iw;
    IntSet::const_iterator         is;
    
    for( t=1; t<=rr; t++ )
    {
	for( p=0; p<n; p++)
	    p2->operator[](p).clear();
	
	if( sk >= 0 )			// there is a sink
	    p1->operator[](sk).clear();
	
	for( p=0; p<n; p++)
	    for( s=0; s<k; s++ )
		foreach_item( iw, p1->operator[](p) )
		    p2->operator[](delta(p,s)).push_back(*iw + char('a'+s));

	if( all || t==rr )
	{
	    res = nil;
	    foreach_item( is, final() ) 
		foreach_item( iw, p2->operator[](*is) )
		    res += new StringAtom( const_cast<char*>((*iw).c_str()));
	    LL += res;
	}
	
	tmp = p1;
	p1  = p2;
	p2  = tmp;
    }

    return reverse( LL );
}

//--------------------------------------------------------------------------------
// eof
