//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 18:20:16 sutner>
// 
// File:     fsmfa.cc
// Purpose:  finite automata (semiautomaton + initial + final)
//
//--------------------------------------------------------------------------------

#include  "stl-wrap.h"
#include  "fsm.h"
#include  "urlist.h"
#include  "listutil.h"
#include  "util.h"

//----------------------------------------------------------------------
//  FA
//----------------------------------------------------------------------

FA::~FA() {}

void  FA::reverseAutomaton( void )
{
   transform( transitions().begin(), transitions().end(), transitions().begin(), 
	      reverse_trans() );   
   IntSet  tmp(initial());
   initial() = final();
   final() = tmp;
}

bool   FA::operator==( const FA& src ) const 
{
   if( &src==this )		// machines identical 
     return 1;

   if( numStates()  != src.numStates()  || 
       numSymbols() != src.numSymbols() || 
       numFinal()   != src.numFinal()   || 
       numInitial() != src.numInitial() )
     return 0;
   
   if( !(transitions()==src.transitions()) )
      return 0;

   if( !(initial()==src.initial()) )
      return 0;

   if( !(final()==src.final()) )
      return 0;

   return 1;
}

//----------------------------------------------------------------------
//  accessible parts 

extern UList  order_select_ulist( UList st, IntSet& reach, int n );

size_t  FA::accessiblePart( FA& ofa, bool norm ) const 
{
    size_t i;
    size_t  n = numStates();
    Graph gr(n);
    IntSet  reach;

    IntSet::const_iterator ritr;
   
    ivector  h(n), hi(n);	// 2.8.1 bug ???
    Fill( h, -1 );
    Fill( hi, -1 );

    toGraph( gr );
    int m = gr.weakComponent( initial(), reach );
    reach.sort();
      
    ofa.set( m, _k );

    int cnt = 0;   
    foreach_item( ritr, reach )      // renumber reachable states  h : R <--> [m] 
	{
	    h[*ritr] = cnt;
	    hi[cnt++] = *ritr;
	}

    TransV  tmp(numTransitions());
    TransV::const_iterator  tit;
   
    size_t nt = 0;
   
    foreach_item( tit, tran )
	if( reach.find(tit->p) && reach.find(tit->q) )
	    tmp[nt++] = *tit;
   
    ofa.transitions().resize( nt );
    for( i=0; i<nt; i++ )
	ofa.transitions()[i] = tmp[i]; 

    ofa.final().initialize(m);		// terrible, fix iterators
    IntSet::const_iterator  itr;

    foreach_item( itr, finst )
	if( reach.find(*itr) )
	    ofa.final().insert( h[*itr] );
   
    ofa.initial().initialize(m);
    IntSet::const_iterator  itrin;
    foreach_item( itrin, initial() )   
	ofa.initial().insert( h[*itrin] );
   

    ofa.accessible() = TRUE;

    if( !norm )
	ofa.states() = order_select_ulist( states(), reach, n );

    return m;
}


//----------------------------------------------------------------------
//  epsilon elimination

size_t   FA::epsilonElimination( FA& mout ) const 
{
    if( eps_free() == TRUE )	// is eps-free	
    {
	mout = *this;
	return numStates();
    }

    size_t  i, n = numStates();
   
    Graph  gr(n);
    toEpsilonGraph( gr );
    vector<IntSet>    clos(n);

    for( i=0; i<n; i++ )
    {
	clos[i].initialize(n);
	gr.weakComponent( i, clos[i] );
    }


    mout.set( n, numSymbols() );
    mout.transitions().resize(0);
    mout.eps_free() = TRUE;

    TransV::const_iterator     tit;
    IntSet::const_iterator  it, iit;
   
    foreach_item( tit, tran )	         // closure for each transition 
	if( tit->s != -1 )
	    foreach_item( it, clos[tit->q] )
		mout.transitions().push_back( Trans( tit->p, tit->s, *it ) );

   
    mout.final() = final();

    mout.initial().initialize(n);	         // closure of the initial states 
    foreach_item( iit, initial() )
	foreach_item( it, clos[*iit] )
	mout.initial().insert(*it);

    return  mout.numTransitions();  
}



//----------------------------------------------------------------------
// power automaton construction, accessible part only
// uses power semimodule algorithm 
//----------------------------------------------------------------------

// currently not const because of check_eps()
void  FA::powerAutomaton( DFA& dfa, bool norm ) 
{
    if( !check_eps() )		// *this is not eps-free
    {
	FA  tfa;
	epsilonElimination( tfa );
	return tfa.powerAutomaton( dfa, norm );
    }
   
    size_t   n = numStates();
    size_t   k = numSymbols();
    UList    stateL = nil;
    IntSet   st(n);

				// build semimodule of pow(Q)
    PSemiModuleGen       pg(n,k);
    delta( pg.next() );
    pg.insert( initial() );
    size_t nn = pg.generate(); 

				// convert into DFA
    dfa.setdfa( nn, k );
    dfa.accessible() = dfa.power() = dfa.eps_free() = TRUE;

				// initial/final states, structured states
    dfa.initial() = 0;

    IntSet::const_iterator  s_itr;
    PSemiModuleGen::iterator  it;
    UList  wits;		
    int    i = 0;

    foreach_item( it, pg )
	{
	    pg.current_set( it, st );

	    bool  isfin = 0;
	    foreach_item( s_itr, st )
		if( final().find( *s_itr ) )
		{
		    isfin = 1; break;
		}
	
	    if( isfin )
		dfa.final().insert(i); 
	
	    if( !norm )
		stateL += intset_to_ulist(st);

	    i++;
	}

    ulist::iterator  lit = pg.delta().begin();
    for( size_t p = 0; p < nn; ++p )
	for( size_t s = 0; s < k; ++s )
	    dfa.delta( p, s ) = *lit++;

    if( !norm )
	    dfa.states() = reverse( substitute( stateL, stateSet() ) );
}




//----------------------------------------------------------------------
// TEMPORARY TEMPORARY TEMPORARY TEMPORARY TEMPORARY TEMPORARY 
//
//  switch to uninit array
//  Note:  must have  n + ceiling(log_2(k)) < 8*sizeof(int) (== 32 usually)
//  here limit to n <= 16

static  int wit_array[1024*64];      // must be initialized to 0
static  int que_array[1024*64];      // no need to initialize

static  size_t  power_size_small( ivector&  tran, 
				  int n, int s, unsigned int initial, size_t& sz )
{
    int *done = que_array;      // points to the front of the search queue 
    int *last = que_array + 1;  // points to the back of the search queue 
    ivector::const_iterator  row;      // points to successive rows of tab 
    int sym = 0;          // current symbol 
    int state = 0;        // current state 
    int next = 0;         // follows state via sym 
    int bitMask = 1;      // successively masks out state bits 
    int i = 0;            // loop counter 

    sz = 1;
    *done = initial;
    wit_array[initial] = INT_MIN | initial;   // high order bit flags a valid element
    while( done < last) 
    {
	state = *done++;
	for( sym = 0, row = tran.begin(); sym < s; sym++ ) 
	{
	    bitMask = 1;
	    next = 0;
	    for( i = 0; i < n; i++) 
	    {
		if (state & bitMask)
		    next |= *row;
		bitMask <<= 1;
		++row;
	    }
	    if( !(wit_array[next] & INT_MIN)) 
	    {
		wit_array[next] = INT_MIN | (sym << n) | state;
		*last++ = next;
		++sz;
	    }
	}
    }
    return sz;
}



// pre:   *this must be eps-free
size_t  FA::powerAutomatonSize() const 
{
    size_t  n = numStates();
    size_t  k = numSymbols();
    size_t  sz;
   
    if( n <= 16 )		// use fast procedure
    {
				// initialize various data structures
	ivector  flatran( k*n );
	Fill( flatran, 0 );
	fill( wit_array, wit_array + int(1U<<n), 0 );

	// translate transitions into flat array
	TransV::const_iterator  it;
	foreach_item( it, tran )
	    flatran[ n* (it->s) + it->p ] |= (1<<(it->q));

	// translate initial states into int
	size_t inist = 0;
	IntSet::const_iterator sit;
	foreach_item( sit, initial() )
	    inist |= (1U << *sit);
       
	power_size_small( flatran, n, k, inist, sz );
    }
    else	// build power aut. using PSemiModuleGen
    {
	PSemiModuleGen   pg(n,k);
	delta( pg.next() );
	pg.insert( initial() );
	sz = pg.generate();
    }
    return sz; 
}

//--------------------------------------------------------------------------------
