//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-26 00:39:45 sutner>
// 
// File:     fsmsa.cc
// Purpose:  base class and semi automata
//
//--------------------------------------------------------------------------------

#include  <queue>

#include  "stl-wrap.h"
#include  "fsm.h"
#include  "urlist.h"
#include  "listutil.h"
#include  "util.h"


//----------------------------------------------------------------------
//  helper functions 

// extern UList  intset_to_ulist( const IntSet& set, int off );

// conversion from flat transition matrix to linear array

static void  copy_triples_to_array( list<Trans>& L, SA& sa )
{
    list<Trans>::iterator  itr;
    foreach_item( itr, L )
	sa.transitionsInsert( itr->p, itr->s, itr->q );
}

// product order, ignore 3rd component
struct  compare_trans_12: public binary_function<bool,Trans,Trans>
{
    compare_trans_12() {}
    bool  operator()( const Trans& S, const Trans& T ) const
    {
	return ( S.p < T.p || ( S.p == T.p && S.s < T.s ) );
    }
};

// product order on states only, ignore symbols
struct  compare_states: public binary_function<bool,Trans,Trans>
{
    compare_states() {}
    bool  operator()( const Trans& S, const Trans& T ) const
    {
	return ( S.p < T.p || ( S.p == T.p && S.q < T.q ) );
    }
};


static  void  sort_transitions( TransV& A, SA::SortOrder how )
{
    if( how == SA::NORMAL_ORDER )
       sort( A.begin(), A.end() );
    else if( how == SA::STATE_ORDER )
       sort( A.begin(), A.end(), compare_states() );
    else
       sort( A.begin(), A.end(), compare_trans() );
}

void   SA::sortTransitions( SA::SortOrder how )
{
    sort_transitions( transitions(), how );
}

//----------------------------------------------------------------------
//  FSMachine
//----------------------------------------------------------------------

FSMachine::~FSMachine() {}

FSMachine& FSMachine::operator=( const FSMachine& src )
{
    if( &src==this )
	return *this;

    _n = src._n;
    _k  = src._k;
    _pr  = src._pr;
    _st  = src._st;

    return *this;
}


//----------------------------------------------------------------------
//  SA
//----------------------------------------------------------------------

SA::~SA() {}

static  bool  is_eps_tran( const Trans& t ) { return  t.s == -1; }

// update properties flags
bool  SA::check_eps()
{
    if( eps_free() == UNDF )
    {
	TransV::iterator it;

	it = find_if( transitions().begin(), transitions().end(), is_eps_tran );
	if( it!=transitions().end() )               // there are eps transitions 
	{	
	    eps_free() = TRUE;
	    return 1;
	}
	eps_free() = FALSE;
	return 0;
    }
    return  eps_free() == TRUE;
}


// reverse all transitions 

void  SA::reverseAutomaton( void )
{
    transform( transitions().begin(), transitions().end(), transitions().begin(), 
	       reverse_trans() );   
    sortTransitions();
}

void  SA::toFA( FA& fa ) const
{
    IntSet  all(_n);
    for( size_t i=0; i<_n; i++ )
	all.insert(i);

    fa.setfa( _n, _k, tran, all, all, properties() );
    fa.sortTransitions();
}

// pre:  transition vector is sorted in standard order
size_t  SA::delta( int p, int s, ilist& next ) const
{
    next.clear();

    Trans  tt(p,s,0);
    TransV::const_iterator  lo, hi, it;
   
    lo = lower_bound( tran.begin(), tran.end(), tt, compare_trans_12() );
    hi = upper_bound( tran.begin(), tran.end(), tt, compare_trans_12() );

    for( it = lo; it != hi; ++it )
	next.push_front(it->q);

    return next.size();
}   

size_t  SA::delta( ilist& curr, int s, ilist& next ) const
{
    IntSet  P(numStates());
    IntSet::iterator  pit;
    
    ilist   tmp;    
    ilist::iterator  it, jt;
    foreach_item( it, curr )
    {
	delta( *it, s, tmp );
	foreach_item( jt, tmp )
	    P.insert( *jt );
    }
    
    next.clear();
    foreach_item( pit, P )
	next.push_back( *pit );
    next.sort();
    
    return  next.size();
}   

// assumes that nextarr has size  n x k and is empty
void  SA::delta( vector<ilist >& nextarr ) const
{
    size_t n = numStates();
    size_t k = numSymbols();
    TransV::const_iterator  it;
    nextarr.resize( n * k );

    foreach_item( it, tran )
	nextarr[ n*(it->s) + it->p ].push_front( it->q );
}   

// SA is assumed to be transitive, then first SCC contains diagonal
// useful only for testing global maps of linear CAs

int  SA::analyzeSCC( void ) const 
{
    Graph  gr;
    SA     sa2;
    size_t  i, n,  nn, m, p0, uu, vv;

    n  = numStates();
    nn = productAutomaton( sa2 );
    sa2.toGraph( gr );

    uvector  scc, card;
    m = gr.stronglyConnectedComponents( scc, card );

    p0 = scc[0];		// the component of the diagonal 

    if( card[p0] > n )		// not surjective 
	return 0;
   
    bool  nontrflag = 0;
    for( i = 0; i < m ; i++ )
	if( i != p0  &&  card[i] > 0 ) { nontrflag = 1; break; }

    if( !nontrflag )		// injective
	return 3;


    Graph coll;			// collapse graph
    coll.initialize( m, 0, 0 );

    ulist::const_iterator it;

    for( i = 0; i < nn; ++i )
    {
	uu = scc[i];
	foreach_item( it, gr[i] )
	    {
		vv = scc[*it];
		if( !coll.edgeQ(uu,vv) && uu != vv)	          // no self loops 
		    coll.insertEdge(uu,vv);
	    }
    }

    // test for openess
    IntSet  reach;
    bool    outflag = 0;

    coll.weakComponent( p0, reach );          // reachable from diagonal
    
    IntSet::const_iterator  itr;
    
    foreach_item( itr, reach )
	if( card[*itr] && *itr != p0 ) { outflag = 1; break; }

    if( outflag )		// not open but surjective 
	return 1;

    Graph rcoll;
    coll.reverse( rcoll );	// reverse of collapse

    rcoll.weakComponent( p0, reach );   // co-reachable from diagonal
    outflag = 0;

    foreach_item( itr, reach )
	if( card[*itr] && *itr != p0 ) { outflag = 1; break; }

    if( outflag )		// not open but surjective 
	return 1;

    return 2;			// open 
}


//----------------------------------------------------------------------
//  epsilon elimination

size_t   SA::epsilonElimination( SA& sout ) const 
{
    if( eps_free() ==  TRUE )	// *this is eps-free
    {
	sout = *this;
	return numTransitions();
    }

    size_t  i, n = numStates();
   
    Graph  gr;
    toEpsilonGraph( gr );
    vector<IntSet>    clos(n);

    for( i=0; i<n; i++ )
    {
	clos[i].initialize(n);
	gr.weakComponent( i, clos[i] );
    }


    sout.set( n, numSymbols() );
    sout.transitions().resize(0);

    TransV::const_iterator   tit;
    IntSet::const_iterator       it;
   
    foreach_item( tit, tran )	         // closure for each transition 
	if( tit->s != -1 )
	    foreach_item( it, clos[tit->q] )
		sout.transitions().push_back( Trans( tit->p, tit->s, *it ) );
    sout.eps_free() = TRUE;

    return sout.numTransitions();  
}


//----------------------------------------------------------------------
//  kernel automaton
//
// input:   SA  *this, initial state set list L = {L1,..,Lr},
//          optional norm
// output:  SA sad, a deterministic SA
//----------------------------------------------------------------------
 

 
size_t  SA::kernelAutomaton( SA& sad, UList& L, bool norm ) const 
{
    size_t       n = numStates(); 
    size_t       k = numSymbols();
    IntSet       st(n);		// current state
    UList        stateL;	// ULists for structured state sets


    PSemiModuleGen   pg(n,k);

    delta( pg.next() );	// precompute next state sets 
   
    UList g;
    while( L )			// initialize generators
    {
	L >> g;
	ulist_to_intset( g, n, st, -1 );
	pg.insert( st );
    }

    pg.generate();		// generate semimodule
   
    if( !norm )			// extract structured statesets 
    {
	PSemiModuleGen::iterator  it;
	foreach_item( it, pg.elements() )
	    {
		pg.current_set( it, st );
		stateL = intset_to_ulist(st) -= stateL;
	    }
    }
    
				// convert into SA
    size_t nn  = pg.size();
   
    sad.set( nn, numSymbols() );
    ulist::iterator  lit = pg.delta().begin();
    for( size_t p = 0; p < nn; ++p )
	for( size_t s = 0; s < numSymbols(); ++s )
	    sad.transitionsInsert( p, s, *lit++ );

    if( !norm )
	if( states() )
	    sad.states() = reverse( substitute( stateL, states() ) );
	else
	    sad.states() = reverse( substitute_int( stateL, n ) );

    return nn;
}


//----------------------------------------------------------------------
//  Welch automaton
//
// input:   SA this
// output:  SA sad, a deterministic SA 
//----------------------------------------------------------------------
 
size_t  SA::welchAutomaton( SA& sad, bool norm ) const 
{
    size_t  n = numStates(); 
    size_t  k  = numSymbols();

    PSemiModuleGen      pg(n,k);

    delta( pg.next() );
   
    IntSet        st(n);		// current state
    UList        sL, stateL;

    // build kernel automaton generated by 0
    st.insert(0);
    pg.insert( st );
    size_t nn = pg.generate();

    // compute max cardinality of new states
    int maxcard = 0;
    size_t  i = 0;
    ivector  cards(nn);
   
    PSemiModuleGen::iterator  it;
    foreach_item( it, pg )
	{
	    pg.current_set( it, st );
	    cards[i++] = st.size();
	}
    // size of Welch automaton
    size_t  nm = 0;
    maxcard = *max_element( cards.begin(), cards.end() );
    count( cards.begin(), cards.end(), maxcard, nm );

    // pi : (nn) --> (nm) partial bijection, inverse of principal
    ivector  pi(nn);
    Fill( pi, -1 );
    int  ii = 0;
    for( i=0; i<nn; i++ )
	if( cards[i] == maxcard )
	    pi[i] = ii++;
   
    sad.set( nm, k );
    sad.transitions().reserve( nm*k );
    sad.transitions().resize(0);

    ulist::iterator  itr = pg.delta().begin();
   
    for( size_t p = 0; p < nm; ++p )
	for( size_t s = 0; s < k; ++s )
	{
	    if( cards[p]==maxcard && cards[*itr]==maxcard )
		sad.transitionsInsert( pi[p], s, pi[*itr] );
	    ++itr;
	}

    // extract max cardinality sets 
    if( !norm )
    {
	UList  stateL;
	i = 0;
	foreach_item( it, pg )
	    if( cards[i++] == maxcard )
	    {
		pg.current_set( it, st );
		stateL += intset_to_ulist( st, 1 );
	    }
	sad.states() = reverse( stateL );
    }
    return nm;
}


//----------------------------------------------------------------------
// conversion to unlabeled graph 
//----------------------------------------------------------------------

bool   SA::toEpsilonGraph( Graph& gr ) const 
{
    bool eps = 0;
    gr.initialize(_n,0,0);
   
    TransV::const_iterator  it;
    foreach_item( it, tran )
	if( it->s == -1 )
	{
	    eps = 1;
	    gr.insertEdge( it->p, it->q );
	}
    return eps;
}


void   SA::toGraph( Graph& gr ) const 
{
    gr.initialize(_n,0,0);
   
    TransV::const_iterator  it;
    foreach_item( it, tran )
	{
	    int p = it->p;
	    int q = it->q;
	    if( !gr.edgeQ(p,q) )
		gr.insertEdge(p,q);
	}
}
   
//----------------------------------------------------------------------
// select the subautomaton described by state set sub
// sub must have same size as input SA 
//----------------------------------------------------------------------

void   SA::select( const IntSet& sub, SA& mout ) const 
{
    assert( numStates() == sub.capacity() );

    list<Trans>  tr;
    TransV::const_iterator  it;

    foreach_item( it, tran )
	if( sub.find(it->p) && sub.find(it->q) )
	    tr.push_front( *it );
   
    mout.set( sub.size(), numSymbols() );
    copy_triples_to_array( tr, mout );
}



void   SA::powerAutomaton( DFA& dfa, bool norm ) const
{
    FA   fa;
    toFA(fa);
    fa.powerAutomaton(dfa,norm);
    return;
}

size_t  SA::powerAutomatonSize() const
{
    FA   fa;
    toFA(fa);
    return  fa.powerAutomatonSize();
}

//----------------------------------------------------------------------
//  product  automaton

static inline int  pairing( int x, int y, int n )
{
    return  x * n + y;
}


// input:  SA this, SA  sa 
// output: sout, the product machine
// 
size_t  SA::productAutomaton( const SA sa, SA& sout ) const 
{
    size_t n1 = numStates();
    size_t n2 = sa.numStates();
    size_t k  = numSymbols();
    size_t i, j, s, nt=0, nn = n1*n2;
   
    TransV  t1(transitions());
    TransV  t2(sa.transitions());
    sort_transitions( t1, MIDDLE_ORDER );
    sort_transitions( t2, MIDDLE_ORDER );

				// compute number of transitions in product
    ivector  trcnt1(k), trcnt2(k);
    Fill( trcnt1, 0 );
    Fill( trcnt2, 0 );

    for( i=0; i<t1.size(); i++ )
	trcnt1[ t1[i].s ]++;

    for( i=0; i<t2.size(); i++ )
	trcnt2[ t2[i].s ]++;

    for( s=0; s<k; s++ )
	nt += trcnt1[s]*trcnt2[s];  

				// build product transitions
    sout.transitions().resize( nt );
    size_t cnt = 0, loi, hii, loj, hij;
   
    hij = hii = 0;
    for( s=0; s<k; s++ )
    {
	loi = hii;
	loj = hij;
	hii += trcnt1[s];
	hij += trcnt2[s];
	for( i=loi; i<hii; i++ )
	    for( j=loj; j<hij; j++ )
	    {
		Trans  &trn = sout.transitions()[cnt++];
		trn.p = pcode( n2, t1[i].p, t2[j].p );
		trn.s = s;
		trn.q = pcode( n2, t1[i].q, t2[j].q );
	    }
    }

    sout.set( nn, k );   
    sout.sortTransitions( NORMAL_ORDER );
    return nn;
}

static inline int  comp_pairing( int n, int x, int y )
{
    int a = ( y < x ) ? y : x;
    int b = ( y > x ) ? y : x;
    return  n * a - ((a*(a+1))/2) + b;
}


// input:  SA this
// output: sout, the product machine *this times *this, condensed form
// n(n+1)/2 states rather than n^2
// 
size_t   SA::productAutomaton( SA& sout ) const 
{
    size_t n = numStates();
    size_t k = numSymbols();
    size_t i, j, s, nt=0, nn = n*(n+1)/2;
   
    TransV  tt(transitions());
    sort_transitions( tt, MIDDLE_ORDER );

				// compute number of transitions in product
    ivector  trcnt(k);
    Fill( trcnt, 0 );

    for( i=0; i<tt.size(); i++ )
	trcnt[ tt[i].s ]++;
   
    for( s=0; s<k; s++ )
	nt += (trcnt[s]*(trcnt[s]+1))/2;  

				// build product transitions
    sout.set( nn, k );   
    sout.transitions().resize( nt );
    size_t  cnt = 0, loi, hii;
   
    hii = 0;
    for( s=0; s<k; s++ )
    {
	loi = hii;
	hii += trcnt[s];
	for( i=loi; i<hii; i++ )    // only ordered pairs (p1,p2) 
	    for( j=i; j<hii; j++ )
	    {
		Trans  &trn = sout.transitions()[cnt++];
		trn.p = comp_pairing( n, tt[i].p, tt[j].p ); 
		trn.s = s;
		trn.q = comp_pairing( n, tt[i].q, tt[j].q ); 
	    }
    }

    return nn;
}


// input:  SA this
// output: sout, the product machine *this times *this, condensed form
// corresponds to  QxQ u Q  
// n(n+1)/2 states rather than n^2
// (p,a,q) and (pp,a,qq)              induces  ((p,pp),a,(q,qq))
// (p,a,q) and not ex qq : (pp,a,qq)  induces  ((p,pp),a,(q,q))
//
// currently not const because of sorting 


size_t   SA::synchProductAutomaton( SA& sout ) const
{
    size_t n = numStates();
    size_t k = numSymbols();
    size_t i, j, nn = n*(n+1)/2;
   
    TransV  tt(transitions());
    sort_transitions( tt, MIDDLE_ORDER );

				// compute number of transitions (_,s,_)
    ivector  trcnt(k);
    Fill( trcnt, 0 );
    TransV::iterator  tr;
    foreach_item( tr, tt )
	trcnt[ tr->s ]++;
				// find all p s.t. there is no (p,a,_) 
				// assumes at least one transition (_,s,_) for each s
    vector<ulist>   gap(k);
    int s = 0;
    int x, pp = -1;
    foreach_item( tr, tt )
    {
	if( tr->s > s )
	{
	    for( x = pp+1; x < int(n); ++x )
		gap[s].push_back( x );
	    s = tr->s;
	    pp = -1;
	}
	
	for( x = pp+1; x < tr->p; ++x )
	    gap[s].push_back( x );
	pp = tr->p;
    }
    for( x = pp+1; x < int(n); ++x )
	gap[s].push_back( x );
   
				// build product transitions
    sout.set( nn, k );   
    size_t  loi, hii;

    hii = 0;
    for( s = 0; s < int(k); s++ )
    {
	loi = hii;
	hii += trcnt[s];
	for( i = loi; i < hii; i++ )    // only ordered pairs (p1,p2) 
	{
	    for( j = i; j < hii; j++ )
		sout.transitionsInsert( comp_pairing( n, tt[i].p, tt[j].p ), s,
					comp_pairing( n, tt[i].q, tt[j].q ) ); 

	    ulist::iterator  pit;
	    foreach_item( pit, gap[s] )
		sout.transitionsInsert( comp_pairing( n, tt[i].p, *pit ), s, 
					comp_pairing( n, tt[i].q, tt[i].q ) );
	}
    }
    
    return nn;
}



III_Triple   SA::toFischer( SA& sa, bool norm ) const
{
    DFA  dfa, mdfa;
    powerAutomaton( dfa, norm );
    size_t n = dfa.minimalAutomaton( mdfa, norm );
    
    int  sk = sink_dfa( mdfa );
    Graph  gr, cgr;
    mdfa.toGraph( gr );
    uvector part, card; 
    gr.makeCollapse( cgr, part, card );
    
    size_t    csk = part[sk];
    ulist::const_iterator it;
    size_t  v;
    for( v = 0; v < n; ++v  )
    {
	bool iscmp = 1;
	foreach_item( it, cgr[v] )
	    if( *it != csk ) { iscmp = 0; break; }
	if( iscmp && v != csk ) break;
    }
    
    size_t  u;
    ulist   L;
    for( u = 0, v = part[v]; u < n; ++u  )
	if( part[u] == v ) L.push_back( u );
    
    SA  smdfa;
    mdfa.toSA( smdfa, norm );
    smdfa.subAutomaton( sa, L, norm );

    return  III_Triple( dfa.numStates(), mdfa.numStates(), sa.numStates() );
}

typedef  hash_map<int,int,hash<int>,equal_to<int> >  iimap;

// extract a subautomaton whose states are given by Q
// need function  h : Q --> (n)   inverse of principal function 
size_t  SA::subAutomaton(  SA& sa, const ulist& Q, bool norm ) const
{
    sa.set( Q.size(), numSymbols() );
    sa.transitions().resize(0);
   
    iimap   h;
    IntSet QQ( numStates() );
    ulist::const_iterator  it;
    int  i = 0;
    foreach_item( it, Q )
    {
	QQ.insert( *it ); 
	h[*it] = i++;
    }
    
    TransV::const_iterator   tit;
   
    foreach_item( tit, transitions() ) // copy proper transitions 
	if( QQ.find( tit->p ) && QQ.find( tit->q ) )
	    sa.transitions().push_back( Trans( h[tit->p], tit->s, h[tit->q] ) );

    if( !norm )
	sa.states() = reverse( substitute( intset_to_ulist( QQ ), stateSet() ) );

    sa.eps_free() = TRUE;
    return sa.numStates();
}


//--------------------------------------------------------------------------------
// pre:    sa is a minimal Fischer automaton
// post:   w is a synchronizing word, returns  Q.w = q

size_t  synchronize_fischer( const SA& sa, string& w ) 
{
    size_t n0 = sa.numStates();
    size_t k  = sa.numSymbols();

				// sp = rev( M x M + M ), compressed representation
    SA  sp;
    sa.synchProductAutomaton( sp );
    sp.reverseAutomaton();
    sp.sortTransitions();

    size_t p, q; 
    ilist  P;
    
    hash_map<size_t,string,hash<size_t>,equal_to<size_t> >  W;
    hash_map<size_t,string,hash<size_t>,equal_to<size_t> >::iterator  it;

    queue<size_t>   Q;
				// push diagonal into Q
    for( p = 0; p < n0; ++p )
    {
	q = comp_pairing( n0, p, p );
	W[q] = "";
	Q.push( q );
    }

    while( !Q.empty() )
    {
	p = Q.front();
	Q.pop();
	for( size_t s = 0; s < k; ++s )
	{
	    sp.delta( p, s, P );
	    while( !P.empty() )
	    {
		q = P.front();
		P.pop_front();
		it = W.find( q );
		if( it == W.end() )
		{
		    W[q] = char('a'+s) + W[p]; // going backwards
		    Q.push(q);
		}
	    }
	}
    }
    
    ilist  SS, S;
    ilist::iterator  sit;
    
    for( p = 0; p < n0; ++p )
	S.push_back(p);
    
    w = "";
    string  u;
    size_t  x;
    
    while( S.size() > 1 )
    {
	sit = S.begin();
	p = *sit++;
	q = *sit;
	x = comp_pairing( n0, p, q );
	u = W[x];
	w = w + u;
 
	for( size_t i = 0; i < u.size(); ++i )
	{
	    sa.delta( S, int( u[i]-'a' ), SS );
	    S = SS;
	}
    }

    assert( S.size()==1 );
    return  S.front();
}

//--------------------------------------------------------------------------------

