//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 15:05:31 sutner>
// 
// File:      graph.cc
// Purpose:   code for some graph algorithms 
//            - Tarjan's SCC algorithm 
//            - collapse of a digraph
//
//--------------------------------------------------------------------------------

#include   "basic.h"

#include   <queue>
#include   "graph.h"
#include   "intset.h"
#include   "ioatom.h"

//--------------------------------------------------------------------------------

// shared globals
static  ivector     dfsnum;
static  ivector     lownum;
static  IntSet        crnt;
static  size_t   scnt, rtn;

//----------------------------------------------------------------------


Graph::~Graph() {}

Graph::Graph(): _nv(0), _ne(0), _nl(0) {}

Graph::Graph( const Graph& src ): 
    _nv(src._nv), _ne(src._ne), _nl(src._nl), adjl(src.adjl), labl(src.labl) {} 

Graph& Graph::operator=( const Graph& src ) 
{
    if( &src != this )
    {
	_nv  = src._nv;
	_ne  = src._ne;
	_nl  = src._nl;
	adjl = src.adjl;
	labl = src.labl;
    }
    return *this;
}



void Graph::initialize( size_t nv, size_t ne, size_t nl )
{
    _nv = nv;
    _ne = ne;
    _nl = nl;
    adjl.resize( nv );
    if( nl ) labl.resize( nv );
}


// Prints graph
ostream&  operator<<( ostream& out, const Graph& gr )
{
    Format fmt;
    
    size_t n = gr.numVertex();
    size_t e = gr.numEdge();
//    int lab = gr.numLabel();
    out << "Graph[" << n << ", " << e << ", 4, " << fmt.lp;
    for ( size_t i=0; i<n-1; i++ )
    {
	Print( out, gr[i], 1 );
	out << ", ";
    }
    Print( out, gr[n-1], 1 );
    out << fmt.rp << "]";
    return out;
}

				// Test for presence of edge(x,y)
bool  Graph::edgeQ( size_t x, size_t y ) const
{
   const_iterator itr;
   foreach_item(itr,adjl[x])
     if( y==*itr ) return 1;
   
   return 0;
}

				// Test for presence of labelled edge(x,y;s)
bool  Graph::edgeQ( size_t x, size_t y , size_t s ) const
{
   const_iterator it1;
   const_iterator it2;
   
   for(it1=adjl[x].begin(),it2=labl[x].begin(); it1!=adjl[x].end(); ++it1,++it2 )
       if( (y==*it1) && (s==*it2) ) return 1;
   return 0;
}

//----------------------------------------------------------------------
// Depth First Search in graph this starting at vertex v

// auxiliary function for DFS
// set reach is not cleared here
void  Graph::dfs_search( size_t v, IntSet& reach ) const
{
    const_iterator  itr;

    foreach_item( itr, adjl[v] )
	if( !reach.find(*itr) )	// (v,w) is tree edge
	{
	    reach.insert(*itr); 
	    dfs_search( *itr, reach ); 		
	}
}


size_t  Graph::weakComponent ( size_t v, IntSet& reach ) const
{
   reach.initialize(_nv);		// clear set reach, init to proper size
   
   reach.insert(v); 
   dfs_search(v,reach);
   
   return reach.size();
}

size_t Graph::weakComponent ( const IntSet& I, IntSet& reach ) const
{
   reach.initialize(_nv);
   
   IntSet::const_iterator itr;
   foreach_item(itr,I)
   {
      reach.insert( *itr );
      dfs_search( *itr, reach );
   }
   
   return reach.size();
}

//----------------------------------------------------------------------
// auxiliary function: compute strongly connected components below v
// add new components to array comps
// cardinality to array card

void  Graph::make_scc( size_t v, uvector& comps, uvector& card ) const
{
    const_iterator itr;
    size_t w = 0, ntcnt;
    lownum[v] = dfsnum[v];
    foreach_item(itr,adjl[v])
	{
	    w = *itr;
	    if( dfsnum[w]==0 )  // (v,w) is tree edge
	    {
		dfsnum[w] = ++scnt;
		crnt.insert(w);
		make_scc( w, comps, card );
		if( lownum[w] < lownum[v] )
		    lownum[v] = lownum[w];
	    }
				// otherwise, update lownum
	    if( dfsnum[w] < dfsnum[v] && 
		crnt.find(w)        && 
		dfsnum[w] < lownum[v] )
		lownum[v] = dfsnum[w];
	}
    
    if( lownum[v]==dfsnum[v] )	// found root v
    {
	ntcnt = 0;		// write SCC, test if non-trivial
	while( !crnt.empty() && 
	       (dfsnum[v] <= dfsnum[crnt.top()]))
	{
	    w = crnt.pop();
	    comps[w] = rtn;
	    ntcnt++;
	}

	if( ntcnt > 1 )		// mark trivial comps 0
	    card.push_back( ntcnt );	      // others with their card
	else
	    card.push_back( edgeQ(w,w) ? 1 : 0 );

	rtn++;
    }
}


//----------------------------------------------------------------------
// compute all strongly connected components of graph
// components are stored in ivector part 
// their cardinalities in ivector card, 0 indicates trivial component

size_t  Graph::stronglyConnectedComponents( uvector& part, uvector& card ) const
{
    size_t  v; 
    scnt = rtn = 0;

    ::initialize( dfsnum, numVertex(), 0 );
    ::initialize( lownum, numVertex(), 0 );
    ::initialize( part, numVertex(), size_t(0) );
    crnt.initialize( numVertex() );

    // generate all strongly connected components
    for( v = 0; v < numVertex(); v++ )
	if( dfsnum[v]==0 )
	{
	    dfsnum[v] = ++scnt;
	    crnt.insert(v);
	    make_scc( v, part, card );
	}
    return rtn;
}


//----------------------------------------------------------------------
// compute collapse graph gc of a directed graph gr 

size_t  Graph::makeCollapse( Graph& gc, uvector& part, uvector& card ) const
{
    size_t m, u, v, uu, vv;

    m = stronglyConnectedComponents(part,card);   

    gc.initialize( m, 0, 0 );
    for( u = 0; u < numVertex(); u++ )    // collapse graph
    {
	const_iterator itr;
	uu = part[u];
	foreach_item( itr, adjl[u] )
        {
	    v  = *itr;
	    vv = part[v];
	    if( !gc.edgeQ(uu,vv) && uu != vv)	          // no self loops 
		gc.insertEdge(uu,vv);
	}
    }

    return m;
}


// Compute reverse of a directed graph 

void  Graph::reverse( Graph& gr ) const
{
    gr.initialize( _nv, _ne, _nl );

    for( size_t u=0; u<_nv; u++ )
    {
	const_iterator itr;
	foreach_item(itr,adjl[u])
	    gr.insertEdge(*itr,u);
    }
}

//----------------------------------------------------------------------
// size of longest path in the acyclic part of a graph, starting at indeg 0 point

size_t  Graph::longestPath() const
{
    size_t u, v, n = numVertex();

    uvector  indeg(n);
    fill( indeg.begin(), indeg.end(), size_t(0) );

    uvector   dist(n);		// max distance to indeg 0 point 
    fill( dist.begin(), dist.end(), size_t(0) );

    const_iterator  vitr;
    queue<size_t,ulist >    Q;

    for( v = 0; v < n; v++ )
	foreach_item( vitr, adjl[v] )
	    indeg[*vitr]++;

    for( v = 0; v < n; v++ )
	if( indeg[v]==0 ) Q.push(v);

    while( !Q.empty() )
    {
	v = Q.front();
	Q.pop();
	foreach_item(vitr,adjl[v])
        {
	    u = *vitr;
	    if( dist[u] < dist[v] + 1 ) dist[u] = dist[v] + 1;
	    if( --indeg[u] == 0 ) Q.push(u);
	}
    }
   
    size_t max = 0;   
    for( v = 0; v < n; v++ )
	if( dist[v] > max ) max = dist[v];
   
    return max;
}

//----------------------------------------------------------------------
// eof



