//----------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 17:43:21 sutner>
// 
// :File:     fsm.h
// :Purpose:  Header file for finite state machines. 
//            
//----------------------------------------------------------------------

#ifndef _FSM_H_
#define _FSM_H_

//----------------------------------------------------------------------

#include  "intset.h"
#include  "pairs.h"
#include  "sgroup.h"
#include  "trees.h"
#include  "ioatom.h"
#include  "urlist.h"		// for structured state sets
#include  "listutil.h"

//----------------------------------------------------------------------

class  FSMachine;		// only states and satellite data
class  SA;			// + transitions
class  FA;			// + initial and final states
class  DFA;			// transition matrix, initial and final states


// property flags for machines

enum  PROPS { FALSE=0, TRUE, UNDF };

struct  Props 
{
    PROPS   _eps;		// has eps-transitions 
    PROPS   _acc;		// is accessible
    PROPS   _pow;		// is a power automaton
    PROPS   _min;		// is minimal DFA
    Props(): _eps(UNDF), _acc(UNDF), _pow(UNDF), _min(UNDF) {}
};



//----------------------------------------------------------------------
// general finite state machines

class  FSMachine 
{
protected:

    size_t  _n;			// number of states 
    size_t  _k;			// input alphabet 0,..,_k-1,  eps is -1 
    UList   _st;		// list for structured state sets 
    Props   _pr;		// additional information 

public:

    FSMachine() {}

    FSMachine( const FSMachine& src ):
	_n(src._n), _k(src._k), _st(src._st), _pr(src._pr) {}

    FSMachine( int ns, int nsm, Props prop=Props(), UList st=nil ):
	_n(ns), _k(nsm), _st(st), _pr(prop) {}

    virtual ~FSMachine();

    size_t   numStates()  const { return _n; }
    size_t   numSymbols() const { return _k; }

   
    UList&  states()          { return _st; }
    UList   states() const    { return _st; }
    UList   stateSet() const  { return _st ? _st : range_i(_n); }

    void  set( int ns, int nsm, Props prp=Props(), UList st=nil  )
    {
	_n = ns; _k = nsm; _pr = prp; _st = st;
    }

    Props    properties()         const { return _pr; }
    virtual  PROPS   eps_free()   const { return _pr._eps; }
    virtual  PROPS   accessible() const { return _pr._acc; }
    virtual  PROPS   minimal()    const { return _pr._min; }
    virtual  PROPS   power()      const { return _pr._pow; }
    Props&   properties()               { return _pr; }
    virtual  PROPS&  eps_free()         { return _pr._eps; }
    virtual  PROPS&  accessible()       { return _pr._acc; }
    virtual  PROPS&  minimal()          { return _pr._min; }
    virtual  PROPS&  power()            { return _pr._pow; }
    virtual  bool    has_states()       { return _st; }


    // undefined 
    FSMachine& operator=( const FSMachine& M );
    FSMachine( FSMachine& M,  UList sub );

    friend ostream& operator<<(ostream& out, const FSMachine& M );   
};				// end class FSMachine


//  int pp,  Machine mm
#define   foreach_state( pp, mm )   for( (pp)=0; (pp)<(mm).numStates(); (pp)++ )
#define   foreach_symbol( ss, mm )  for( (ss)=0; (ss)<(mm).numSymbols(); (ss)++ )


//----------------------------------------------------------------------
// semi-automata 
//----------------------------------------------------------------------



class  SA: public FSMachine
{
protected:

    TransV   tran;

public:

    SA( int ns=0, int nsm=0, Props prp=Props(), UList stl=nil ):
	FSMachine(ns,nsm,prp,stl)  { tran.resize(0); }

    SA( int ns, int nsm, const TransV& tr, Props prp=Props(), UList stl=nil ):
	FSMachine(ns,nsm,prp,stl), tran(tr) { check_eps(); }

    SA( const SA& M ): FSMachine(M._n,M._k,M._pr,M._st), 
	tran(M.tran)  { check_eps(); }

    virtual ~SA();
    
    bool  check_eps();		// return true if there are eps-moves, update Props

    int   numTransitions() const { return tran.size(); }

          TransV&  transitions()       { return tran; } 
    const TransV&  transitions() const { return tran; } 

    void  transitionsInsert( int p, int s, int q )   { tran.push_back(Trans(p,s,q)); }
//    void  transitionsInsert( const Trans& tt )  { tran.push_back(tt); }
   
   
    void  setsa( int ns, int nsm, const TransV& tr, Props prp=Props(), UList stl=nil )
    {
	set( ns, nsm, prp, stl );
	tran = tr;
    }
   
    size_t   delta( int p, int s, ilist& next ) const;
    size_t   delta( ilist& curr, int s, ilist& next ) const;
    void     delta( vector<ilist >& nextarr ) const;

    void     select( const IntSet& sub, SA& mout ) const;

    enum     SortOrder { NORMAL_ORDER, MIDDLE_ORDER, STATE_ORDER };
    void     sortTransitions( SortOrder how = NORMAL_ORDER );

    void     toFA( FA& fa ) const;

    size_t   kernelAutomaton( SA& mout, UList& L, bool norm=1 ) const;
    void     powerAutomaton(  DFA& mout, bool norm=1 ) const;
    size_t   subAutomaton(  SA& sa, const ulist& Q, bool norm=1 ) const;
    size_t   powerAutomatonSize() const;
    size_t   productAutomaton( const SA sa, SA& sout ) const;  // product this and sa
    size_t   productAutomaton( SA& sout ) const;   // square of this, condensed
    size_t   synchProductAutomaton( SA& sout ) const;
    int      analyzeSCC() const;
    void     reverseAutomaton();
    size_t   welchAutomaton( SA& sa, bool norm ) const;
    size_t   epsilonElimination( SA& sout ) const;
    void     eps_closure( const Graph&, vector<IntSet>&, list<Trans>& ) const;
    void     toGraph( Graph& gr ) const;
    bool     toEpsilonGraph( Graph& gr ) const;
    III_Triple   toFischer( SA& sa, bool norm=1 ) const;

    friend ostream& operator<<(ostream& out, const SA& M );

};				// end class SA


//----------------------------------------------------------------------

class  FA: public SA
{
    IntSet  inist;		// initial states 
    IntSet  finst;		// final states 

public:

    FA( int ns=0, int nsm=0, Props prp=Props(), UList stl=nil ): 
	SA(ns,nsm,prp,stl), inist(ns), finst(ns)  {}

    FA( int ns, int nsm, const TransV& tr, Props prp=Props(), UList stl=nil ): 
	SA(ns,nsm,tr,prp,stl), inist(ns), finst(ns)  {}

    FA( int ns, int nsm, const TransV& tr, IntSet& ini, IntSet& fin, 
	Props prp=Props(), UList stl=nil ): 
	SA(ns,nsm,tr,prp,stl), inist(ini), finst(fin)  {}

    FA( const FA& M ):
	SA( M._n, M._k, M.tran, M._pr, M._st ), 
	inist(M.inist), finst(M.finst)  {}

//   FA( FA& M,  UList sub ) {}

    ~FA();
    

    size_t          numInitial() const { return inist.size(); }
    size_t          numFinal()   const { return finst.size(); }
    IntSet&         initial()          { return inist; } 
    IntSet&         final()            { return finst; } 
    const IntSet&   initial()    const { return inist; } 
    const IntSet&   final()      const { return finst; } 

    void  setfa( int ns, int nsm, const TransV tr, 
	       const IntSet& in, const IntSet& fi, Props prp=Props(), UList stl=nil )
    {
	setsa( ns, nsm, tr, prp, stl );
	initial()    = in;
	final()      = fi;
    }
   
    size_t   accessiblePart( FA& mout, bool norm=1 ) const;
    size_t   epsilonElimination( FA& mout ) const;
    size_t   kernelAutomaton( FA& mout, UList& L, bool norm=1 ) const;
    void     reverseAutomaton();
    void     powerAutomaton(  DFA& dfa, bool norm=1 );
    size_t   powerAutomatonSize() const;

    bool  operator==( const FA& M ) const; // equality testing for FAs

    friend ostream& operator<<(ostream& out, const FA& M );
   
};				// end class FA

//----------------------------------------------------------------------

class  DFA: public FSMachine
{
    vector<size_t>   trmat;		// flat kxn transition matrix
    size_t     inist;		// initial state
    IntSet     finst;		// final states 

public:

    // only sizes known, no transitions/final states
    DFA( int ns=0, int nsm=0 );

    DFA( const DFA& );

    ~DFA();

    // reset data in DFA
    void setdfa( int n, int k, Props prp = Props(), UList sts=nil )
    {
	set( n, k, prp, sts );
	trmat.resize( n*k );
	finst.initialize(n);
    }
   

    size_t&    initial()        { return inist; } 
    size_t     initial()  const { return inist; } 
    size_t     numFinal() const { return finst.size(); }
    IntSet&       final()          { return finst; } 
    const IntSet& final() const    { return finst; } 
    vector<size_t>&        transitions()       { return trmat; } 
    const vector<size_t>&  transitions() const { return trmat; } 
    size_t&       delta( size_t p, size_t s )       { return trmat[_n*s+p]; }
    size_t        delta( size_t p, size_t s ) const { return trmat[_n*s+p]; }

    void    toSA( SA& sa, bool norm=1 ) const;
    void    toGraph( Graph& gr ) const;
    size_t  accessiblePart( DFA& odfa, bool norm=1 ) const;
    size_t  minimalAutomaton( DFA& mout, bool norm=1 ) const;
    size_t  productAutomaton( DFA& M, DFA& mout, bool un, bool norm ) const;
    size_t  quotientAutomaton( DFA& M, FA& fa, bool norm ) const;
    void    reverseAutomaton( FA& mout ) const;
    bool    operator==( const DFA& M ) const; // equality testing for DFAs
   
    UList language( int rr ) const;

    friend ostream& operator<<(ostream& out, const DFA& M );
};				//  end class DFA 


int  sink_dfa( const DFA& dfa );

//----------------------------------------------------------------------
//  BEFAs and rational expressions 
//
// implicitely begin = 0, exit = _n-1
// no transitions to begin, from exit
// stored in flat array  p0 s0 q0 p1 s1 q1 ... 
// 0 <= p_i,q_i < _n  and  s_i is the int value of a char 
//----------------------------------------------------------------------
 
class  BEFA
{
private:

    int       _n;			// number of states 
    ivector  _tran;			// flat array of transitions


public:

    BEFA(): _n(-1) {}

    explicit  BEFA( int r );	        // special machines: eps (0), empty (-1) 
    BEFA( char *ww, char off='a' );     // machine for one word ww, _n = |ww|
    ~BEFA();
   
    BEFA( const BEFA& M );   
    BEFA& operator=( const BEFA& M );   
    int   operator==( const BEFA& M ) const;   


    void initialize( int n, int t );	// _n = n, allocate t transitions

    int  numStates()  const     { return _n; }
    int  numTransitions() const { return _tran.size()/3; }
    ivector&       transitions()       { return _tran; }
    const ivector& transitions() const { return _tran; }

    // (this,m2) -> mout
    // always return size of result
    int   catenation( const BEFA& M, BEFA& mout ) const;
    int   sum( const BEFA& M, BEFA& mout ) const;
    int   star( BEFA& mout ) const;
    int   convert( FA *mout ) const;

    friend  ostream& operator<<( ostream& out, const BEFA& M );
};


//----------------------------------------------------------------------
// free functions 

void   IthSymbolFA( int a, int n, int k, FA& fa );
void   make_basic_dfa( int k, int tt, DFA& dfa );

size_t  synchronize_fischer( const SA& sa, string& w );

//----------------------------------------------------------------------
#endif
