//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:40:54 sutner> 
//
//  :File:      ioatom.h
//  :Purpose:   writing transitions, matrices, lists to ostream
//              used mostly in ioatom.cc 
//---------------------------------------------------------------------------

#ifndef _PRINT_H_
#define _PRINT_H_

//----------------------------------------------------------------------

#include  <iostream>
#include  <list>
#include  <vector>
#include  <algorithm>

#include  "basic.h"
#include  "urlist.h"
#include  "fsm.h"
#include  "intset.h"
#include  "pairs.h"
#include  "stl-wrap.h"

//--------------------------------------------------------------------------------
// file I/O

UList  read_atom( UList );
UList  write_atom( UList, UList );

UList  read_atom_raw( istream& in );
void   write_atom_raw( ostream& os, UList A, bool norm );
void   write_atom_raw( ostream& os, const SA& sa, bool norm=1 );
void   write_atom_raw( ostream& os, const FA& fa, bool norm=1 );
void   write_atom_raw( ostream& os, const DFA& dfa, bool norm=1 );
void   write_atom_raw( ostream& os, const CA& dfa, bool norm=1 );
void   write_atom_raw( ostream& os, const SGroup& sg, bool norm );

//----------------------------------------------------------------------
// general printing for containers
// formats into lines with ww items each
// defaults:
//    ww  = -1           reset to len

template<class Iter>
void  PrintPlain( ostream& out, Iter first, int len, int ww=-1 )
{
    Format  fmt;
    
    if( len==0 )
    {
	out << fmt.lp << fmt.rp;
	return;
    }
    
    int  i, j;
    if( ww==-1 )
	ww = len;
    int  nrows = len/ww;
    int  rem = len - (nrows-1)*ww;

    out << fmt.lp;
    for( i=0; i<nrows-1; i++ )
    {
	for( j=0; j<ww-1; j++ )
	{
	    out << *first++;
	    out << fmt.cm;
	}
	out << *first++;
	out << fmt.cm << endl;
    } 
    for( j=0; j<rem-1; j++ )
    {
	out << *first++;
	out << fmt.cm;
    }
    out << *first++;
    out << fmt.rp;
}


//----------------------------------------------------------------------
// print container in matrix format
// dimensions: len/ww  by  ww 
// this assumes  len  is a multiple of ww, otherwise truncates len % ww items
// defaults:
//    ww  = -1           reset to len
//    beg = "{"
//    end = "}"


template<class Iter>
void  PrintMatrix( ostream& out, Iter first, int len, int ww=-1 ) 
{
    Format  fmt;

    if( len==0 )
    {
	out << fmt.lp << fmt.rp;
	return;
    }
    int  i, j;
    if( ww==-1 ) ww = len;
    int  nrows = len/ww;

    out << fmt.lp;
    for( i=0; i<nrows-1; i++ )
    {
	out << fmt.lp;	
	for( j=0; j<ww-1; j++ )
	{
	    out << *first++;
	    out << fmt.cm;
	}
	out << *first++;
	out << fmt.rp << fmt.cm;
    } 
    out << fmt.lp;	
    for( j=0; j<ww-1; j++ )
    {
	out << *first++;
	out << fmt.cm;
    }
    out << *first++;
    out << fmt.rp << fmt.rp;
}


//----------------------------------------------------------------------
// manipulator versions
// apply a formatting function ff to each item before printing 
//----------------------------------------------------------------------

template<class Iter,class Manip>
void  PrintManip( ostream& out, Iter first, int len, Manip ff, int ww=-1 )
{
    Format  fmt;

    if( len==0 )
    {
	out << fmt.lp << fmt.rp;
	return;
    }

    int  i, j;
    if( ww==-1 ) ww = len;
    int  nrows = len/ww;
    int  rem = len - (nrows-1)*ww;

    out << fmt.lp;
    for( i=0; i<nrows-1; i++ )
    {
	for( j=0; j<ww-1; j++ )
	{
	    out << ff(*first++);
	    out << fmt.cm;
	}
	out << ff(*first++);
	out << fmt.cm << endl;
    } 
    for( j=0; j<rem-1; j++ )
    {
	out << ff(*first++);
	out << fmt.cm;
    }
    out << ff(*first++);
    out << fmt.rp;
}


//----------------------------------------------------------------------

template<class Iter, class Manip>
void  PrintMatrixManip( ostream& out, Iter first, int len, Manip ff, int ww=-1 )
{
    Format  fmt;

    if( len==0 )
    {
	out << fmt.lp << fmt.rp;
	return;
    }

    int  i, j;
    if( ww==-1 ) ww = len;
    int  nrows = len/ww;

    out << fmt.lp;
    for( i=0; i<nrows-1; i++ )
    {
	out << fmt.lp;	
	for( j=0; j<ww-1; j++ )
	{
	    out << ff(*first++);
	    out << fmt.cm;
	}
	out << ff(*first++);
	out << fmt.rp << fmt.cm;
    } 
    out << fmt.lp;	
    for( j=0; j<ww-1; j++ )
    {
	out << ff(*first++);
	out << fmt.cm;
    }
    out << ff(*first++);
    out << fmt.rp << fmt.rp;
}


//----------------------------------------------------------------------
// manipulators
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// sequences

inline void  Print( ostream& out, const ulist &L, int off=0, int ww=-1 )
{
    PrintManip( out, L.begin(), L.size(), bind2nd( plus<int>(), off ), ww );
}

inline void  Print( ostream& out, const ivector &A, int off=0, int ww=-1 )
{
    PrintManip( out, A.begin(), A.size(), bind2nd( plus<int>(), off ), ww );
}

//----------------------------------------------------------------------
// transitions, really should use an alphabet 


struct  iii_to_ici: public binary_function<III_Triple,int,ICI_Triple>
{
    iii_to_ici() {}

    ICI_Triple  operator()( const III_Triple& t, int off ) const
    {
	int p,s,q;
	t.get(p,s,q);
	Triple<int,char,int>   tt;
	if(s==-1)
	    tt.set(p+off,'@',q+off);
	else
	    tt.set(p+off,char(s+'a'),q+off);
	return tt;
    }


};


// internal symbols: -1 <= s <  k
// external symbols:  0 <= s <= k
// internal states:   0 <= s <  n
// external states:   1 <= s <= n

struct  shift_trans: public binary_function<Trans,int,Trans>
{
    shift_trans() {}

    Trans  operator()( const Trans& T, int off ) const
    {
	return  Trans( T.p+off,T.s+1,T.q+off );
    }
};


inline void  PrintTransitions( ostream& os, const TransV& arr, int off=0, int ww=-1 )
{
    PrintManip( os, arr.begin(), arr.size(), 
		bind2nd( shift_trans(), off ), ww );
}

//----------------------------------------------------------------------
// IntSet

inline void  Print( ostream& out, const IntSet& S, int off=0, int ww=-1 )
{
    vector<int>   A(S.size());
    IntSet::const_iterator it;
    int i=0;
    foreach_item( it, S )
	A[i++] = *it + off;
	
    PrintPlain( out, A.begin(), A.size(), ww );
}



//----------------------------------------------------------------------
#endif






