//--------------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:50:17 sutner>
// 
//  :File:     trees.h
//  :Purpose:  various types of tries
//
//--------------------------------------------------------------------------------

#ifndef _TREES_H
#define _TREES_H


//----------------------------------------------------------------------
#include  <assert.h>
#include  <iostream>

#include  "basic.h"
#include  "stl-wrap.h"
#include  "intset.h"
#include  "pairs.h"
#include  "list.h"
#include  "map.h"
#include  "algo.h"

//----------------------------------------------------------------------
// classes defined herein

class  bTrie;
class  bTrieNode;
class  nTrie;
class  nTrieNode;

typedef   bTrieNode   *bTrieP;
typedef   nTrieNode   *nTrieP;

//----------------------------------------------------------------------
//  binary tries
//  uses memory manager 
//----------------------------------------------------------------------

class  bTrieNode
{
    friend class bTrie;

    bTrieNode  *lptr, *rptr, *pptr; // left/right son, parent
    int         lrson;		    // left or right son indicator
    int         aux;		    // auxiliary 

public:

    bTrieNode(): lptr(NULL), rptr(NULL), pptr(NULL), lrson(-1), aux(-1) {}

    // append node in position lr
    bTrieNode( bTrieNode *par, int lr, int x ):
	lptr(NULL), rptr(NULL), pptr(par), lrson(lr), aux(x) {}
   
    ~bTrieNode();			  // remove whole subtree

    bTrieP  parent()    { return pptr; } 
    int     whichSon()  { return lrson; }
    int     content()   { return aux; }
				// use aux for levels
    bTrieP  insert( int lr );
};

//----------------------------------------------------------------------

class  bTrie
{
    bTrieP       root;		// the root of the trie
    size_t       dpth;		// depth 
    size_t       numbr;		// number of branches 
   
public:

    bTrie(): root(NULL), dpth(0), numbr(0) {}
 
    bTrie( int d ): dpth(d), numbr(0) { root = new bTrieNode(NULL,0,0); }
   
    ~bTrie();
   
    int  size() const { return numbr; }
    int  depth()  const { return dpth; }

// check if there, insert if not  
    int  probe_branch( const IntSet& bitv, bTrieP& leaf );
    int  probe_branch( int *ip, bTrieP& leaf );	// needed ????
    
};

// note: set must have proper capacity, and must be cleared 
void  retrieve_bbranch( bTrieP btn, IntSet& st );


//----------------------------------------------------------------------
//  n-ary tries 
// need pointer to various things at the leafs, uses generic pointers
// uses memory manager
//----------------------------------------------------------------------



class  nTrieNode
{
    friend class nTrie;

    int             whichson;	// position as son of father 
    nTrieP          par;	// pointer to father 
    vector<nTrieP>  sons;	// pointers to sons 
    int             aux;	// running counter

public:

    nTrieNode(): whichson(0), par(NULL), aux(0) {}
   
    nTrieNode( int n ): whichson(0), par(NULL), sons(n,nTrieP(NULL)), aux(0) {}
   
    // append node to p in position ii
    nTrieNode( int w, nTrieP  p, int ii, int aa=0 ): 
	whichson(ii), par(p), sons( w, nTrieP(NULL) ), aux(aa)
    {
	if( par ) par->sons[ii] = this;
    }
   
    ~nTrieNode();			  // remove whole subtree

    int       width() const      { return sons.size(); }
    int       whichSon() const   { return whichson; }
    int&      content()          { return aux; }
    nTrieP    parent()           { return par; }

    nTrieP  operator[]( int i ) { return sons[i]; }
    
    // insert at current node
    nTrieP  insert( int ii )
    {
	return sons[ii] = new nTrieNode( width(), this, ii );
    }

};


//----------------------------------------------------------------------


class  nTrie
{
    size_t        wdth;		// width (max. branching at each node)
    size_t        dpth;		// depth 
    size_t        numbr;	// number of branches 
    nTrieP        root;		// the root of the trie
   
public:

    nTrie(){}
 
    nTrie( int w, int d ): 
	wdth(w), dpth(d), numbr(0), root( new nTrieNode(w,NULL,0) ) {}
   
    ~nTrie();
   
    size_t&     size()     {return numbr;}
    size_t&     depth()    {return dpth;}
    size_t&     width()    {return wdth;}

    // return 1 if branch is already there, otherwise insert and return 0
    // set lf to point to leaf of branch 
    bool  probe_branch( uvector::iterator tt, nTrieP& lf );
};

// traverse from leaf back to root, write to tt 
void  retrieve_branch( const nTrieP& leaf, uvector& tt );
void  retrieve_branch_reverse( const nTrieP& leaf, uvector::iterator& it, int n );

//----------------------------------------------------------------------
#endif  // _TREES_H
