//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 14:13:33 sutner> 
//
// :File:      urlist.h
// :Purpose:   Rich handle and atom classes,
//             implements nested lists of urelements. 
// 
//---------------------------------------------------------------------------


#ifndef  _URLIST_H_
#define  _URLIST_H_

#include  "basic.h"
#include  "Clock.h"
#include  "integer.h"
#include  <string>
#include  <algorithm>
#include  <alloc.h>


// bad hack, should have initialization procedure 
#define  nil  UList()

//--------------------------------------------------------------------------------


struct  UList;
struct  Atom;
struct  NilAtom;
struct  IntAtom;
struct  StringAtom;
struct  ConsAtom;
struct  RefAtom;
struct  UListInitializer;


typedef  Atom           *AtomP;
typedef  const UList&   CRef;

// include useratoms
enum  AType { NIL=0, FUNC, CONS, INT, STR, IARR, INTL, SAA, FAA, DFAA, CAA, SGA, GRA,
ATEXN, VEXN, AEXN, FCTN, FEXN, UFCTN, UFCTE, DEFN, STMN, IFN, WHLN, ASSA, COMPN };

class  FSMachine;
class  SA;
class  FA;
class  DFA;
class  CA;
class  Graph;
class  SGroup;

class  SymTable;

//--------------------------------------------------------------------------------


struct  UList 
{
    typedef  __simple_alloc<ConsAtom,std::__single_client_alloc> cons_alloc;
    typedef  __simple_alloc<IntAtom,std::__single_client_alloc> int_alloc;
    
    friend struct UListInitializer;
    static  void  initialize();
    static  void  finalize();

public:

    UList();
    UList( const Atom* );
    UList( CRef, CRef );

    UList( CRef );
    UList&  operator=( CRef );
    ~UList();

    UList    copy();
    UList    clone() const;
    void     destroy();

    UList    head() const;
    UList    tail() const;
    operator bool() const; 
    UList    operator*();
    UList    operator++();
    UList&   operator+=( CRef );
    UList&   operator+=( int n );
    UList&   operator+=( const char* s );

    AType    atom_type() const;
    bool     operator==( CRef );
    bool     operator<( CRef );
    void     print( ostream& ) const;
    void     sub_print( ostream& ) const;

    UList    eval( SymTable& );

    UList    operator()() const;
    UList    operator()(UList) const;
    UList    operator()(UList,UList) const;
    UList    operator()(UList,UList,UList) const;

    friend   int&  ivalue( CRef );
    friend   char *svalue( CRef );
    friend   Integer&  Ivalue( CRef a );  
    friend   vector<int>&  iavalue( CRef a ); 
    friend   SA&   savalue( CRef a );     
    friend   FA&   favalue( CRef a );     
    friend   DFA&  dfavalue( CRef a );    
    friend   CA&   cavalue( CRef a );     
    friend   Graph&  grvalue( CRef a );   
    friend   SGroup&  sgvalue( CRef a );  

    friend  ostream&  operator<<( ostream&, CRef );
    friend  int refcount( CRef );
    friend  int arity( CRef );
    friend  UList  Head( UList );
    friend  UList  Tail( UList );
    friend  AtomP ufunc( CRef );

private:
    AtomP     ap;
};


struct  UListInitializer
{
    Clock  ticktock;
    static int ulinitcnt;
    static bool showtime;
    
    UListInitializer() 
    { 
	if( !ulinitcnt++ ) UList::initialize(); 
    }
    ~UListInitializer() 
    { 
	if( showtime )
	    cerr << "*** time: " << ticktock.lapTime() << endl; 
#ifdef  TRCBUG
	TraceCensus(1);
#endif
	if( !--ulinitcnt ) UList::finalize(); 
    }
};



class  Atom
{
    friend struct UList;
public:
    virtual  ~Atom();
    virtual  AtomP        destroy();
    virtual  const Atom  *copy();
    virtual  const Atom  *clone() const;
    virtual  UList   head() const;
    virtual  UList   tail() const;
    virtual  AType   atom_type() const = 0;
    virtual  void    sub_print( ostream& os ) const;
    virtual  void    print( ostream& ) const;
    virtual  UList   eval( SymTable& );
    virtual  UList  operator()() const;
    virtual  UList  operator()(UList) const;
    virtual  UList  operator()(UList,UList) const;
    virtual  UList  operator()(UList,UList,UList) const;
};

//--------------------------------------------------------------------------------
//  base atoms: nil, function, ref counted

struct NilAtom: public Atom
{
    ~NilAtom();
    NilAtom  *destroy();
    const NilAtom  *copy();
    const NilAtom  *NilAtom::clone() const;
    AType     atom_type() const;
    void  print(ostream&) const;
    void  sub_print(ostream&) const;
    UList    eval( SymTable& );
};


//--------------------------------------------------------------------------------
// ref counted atoms

struct RefAtom: public Atom
{
//    Trace trc;
    int   rc;

    RefAtom();
    RefAtom         *destroy();
    const  RefAtom  *copy();
};


//--------------------------------------------------------------------------------
// ground ref counted atoms: cons, int, string 


struct IntAtom: public RefAtom
{
    int  val;
    ~IntAtom();
    IntAtom( int n=0 );
    IntAtom  *destroy();
    const IntAtom  *clone() const;
    AType     atom_type() const;
    int&      value();
    const int value() const;
    void  print(ostream&) const;
    void  sub_print(ostream&) const;
    UList    eval( SymTable& );
//    UList   head() const;	// permissive semantics for extraction
};

struct  StringAtom: public RefAtom
{
    char *str;
    StringAtom( const char* ss ="" );
    StringAtom( const string& );
    ~StringAtom();
    const StringAtom  *clone() const;
    AType     atom_type() const;
    char     *value();
    const char *value() const;
    void  print(ostream&) const;
    void  sub_print(ostream&) const;
    UList    eval( SymTable& );
};

struct  ConsAtom: public RefAtom
{
    UList  car, cdr;

    ConsAtom();
    ~ConsAtom();
    ConsAtom( const ConsAtom& );
    ConsAtom( CRef, CRef );
    UList       head() const;
    UList       tail() const;
    ConsAtom       *destroy();
    const ConsAtom *clone() const;
    AType       atom_type() const;
    void        print(ostream&) const;
    void        sub_print(ostream&) const;
    UList       eval( SymTable& );
};



//--------------------------------------------------------------------------------
// allocator for int/cons cells
//--------------------------------------------------------------------------------

inline  ConsAtom  *new_cons( CRef a, CRef b )
{
#ifndef  ALLOCATOR
    return new ConsAtom(a,b);
#else
    ConsAtom *pp = UList::cons_alloc::allocate();
    construct( pp, ConsAtom() );
    construct( &pp->car, a );
    construct( &pp->cdr, b );
    return  pp;
#endif
}

inline  void delete_cons( ConsAtom *pp )
{
#ifndef  ALLOCATOR
    delete pp;
#else
    destroy( &pp->car );
    destroy( &pp->cdr );
    UList::cons_alloc::deallocate( pp );
#endif
}


inline  IntAtom  *new_int( int x )
{
#ifndef  ALLOCATOR
    return new IntAtom(x);
#else
    IntAtom *pp = UList::int_alloc::allocate();
    construct( pp, IntAtom() );
    construct( &pp->val, x );
    return  pp;
#endif
}

inline  void delete_int( IntAtom *pp )
{
#ifndef  ALLOCATOR
    delete pp;
#else
    destroy( &pp->val );
    UList::int_alloc::deallocate( pp );
#endif
}

//--------------------------------------------------------------------------------
// free list operations
//--------------------------------------------------------------------------------

ostream&  operator<<( ostream& os, AType tt );

inline  bool   isTrue( UList L ) 
{
    if( L.atom_type() == INT ) 
	return  ivalue( L );
    return L;
}

// cons, both prefix and infix (right associative)

inline  UList  cons( CRef a, CRef b )       { return new_cons(a,b); }
inline  UList  operator-=( CRef a, CRef b ) { return new_cons(a,b); }

inline  UList  operator-=( int n, CRef b ) 
{ 
    return new_cons( new_int(n), b ); 
}

inline  UList  operator-=( const string& s, CRef b ) 
{ 
    return new_cons( new StringAtom(s), b ); 
}

inline  UList&  UList::operator+=( int n )
{
    return  this->operator+=( new_int(n) );
}

inline  UList&  UList::operator+=( const char *s )
{
    return  this->operator+=( new StringAtom(s) );
}

//--------------------------------------------------------------------------------
// ground atoms 

#ifdef URLBUG
inline  int&  ivalue( CRef a ) 
{ 
    IntAtom* p = dynamic_cast<IntAtom*>(a.ap);
    if( p ) return p->value(); 
    Error( "ivalue: wrong atom" );
    exit(1);
}
inline  char *svalue( CRef a ) 
{ 
    StringAtom* p = dynamic_cast<StringAtom*>(a.ap);
    if( p ) return p->value(); 
    Error( "svalue: wrong atom" );
    exit(1);
}
#else
inline  int&  ivalue( CRef a ) { return ((IntAtom*)(a.ap))->value(); } 
inline  char *svalue( CRef a ) { return ((StringAtom*)(a.ap))->value(); }
#endif


inline  UList  toul( const string& s ) { return new StringAtom(s);}
inline  UList  toul( int n )           { return new_int(n); }


//--------------------------------------------------------------------------------
// extraction ops

inline  UList& operator>>( UList& src, UList& targ )
{
    targ = *src;
    ++src;
    return src;
}

inline  UList& operator>>( UList& src, int& targ )
{
    targ = ivalue(*src);
    ++src;
    return src;
}

inline  UList& operator>>( UList& src, string& targ )
{
    targ = svalue(*src);
    ++src;
    return src;
}

inline  UList  Head( UList L ) { return L.head(); }
inline  UList  Tail( UList L ) { return L.tail(); }

//--------------------------------------------------------------------------------
//  lots of inlines
//--------------------------------------------------------------------------------

extern  NilAtom   *_nilatomptr_;

inline  UList::UList(): ap(_nilatomptr_) {}
inline  UList::UList( const Atom* aptr ): ap(const_cast<AtomP>(aptr)) {}
inline  UList::UList( CRef src ): ap(const_cast<AtomP>(src.ap->copy())) {}
inline  UList  UList::head() const { return ap->head(); }
inline  UList  UList::tail() const { return ap->tail(); }
inline  UList::operator bool() const { return ap != _nilatomptr_; }
inline  UList  UList::operator*() { return head(); }
inline  UList  UList::copy() { return  ap->copy(); }
inline  UList  UList::clone() const { return  ap->clone(); }
inline  UList::~UList() { delete  ap->destroy(); } 
inline  UList  UList::operator()() const { return ap->operator()(); }
inline  UList  UList::operator()(UList x) const { return ap->operator()(x); }
inline  UList  UList::operator()(UList x,UList y) const { return ap->operator()(x,y); }
inline  UList  UList::operator()(UList x,UList y,UList z) const 
{ return ap->operator()(x,y,z); }
inline  UList  UList::eval( SymTable& env ) { return ap->eval(env); }
inline  AType  UList::atom_type() const { return ap->atom_type(); }	
inline  void  UList::print( ostream& os ) const { ap->print(os); }
inline  void  UList::sub_print( ostream& os ) const { ap->sub_print(os); }

inline  UList& UList::operator=( CRef src )
{
	if( this != &src )
	{
	    delete  ap->destroy();
	    ap = const_cast<AtomP>(src.ap->copy());
	}
	return *this;
}

inline  UList  UList::operator++() 
{ 
    this->operator=( tail() ); 
    return *this; 
}

inline  UList&   UList::operator+=( CRef a )
{ 
    this->operator=( new_cons(a,*this) );
    return *this;
}


//--------------------------------------------------------------------------------
// Atom 

inline  Atom::~Atom() {}
inline  UList   Atom::head() const  { return nil; } 
inline  UList   Atom::tail() const  { return nil; }
inline  AtomP   Atom::destroy()     { return NULL; }
inline  const Atom  *Atom::copy()   { return NULL; }
inline  const Atom  *Atom::clone() const  { return NULL; }
inline  void    Atom::print( ostream& os ) const { os << "*** printing Atom"; }
inline  void    Atom::sub_print( ostream& os ) const { os << " . "; }
inline  UList   Atom::eval( SymTable& ) { return nil; }
inline  UList   Atom::operator()() const { return nil; } 
inline  UList   Atom::operator()(UList) const { return nil; } 
inline  UList   Atom::operator()(UList,UList) const { return nil; } 
inline  UList   Atom::operator()(UList,UList,UList) const { return nil; } 

//--------------------------------------------------------------------------------
// NilAtom

inline  NilAtom::~NilAtom() {}
inline  NilAtom   *NilAtom::destroy() { return NULL; }
inline  const NilAtom  *NilAtom::copy()  { return _nilatomptr_; }
inline  const NilAtom  *NilAtom::clone() const { return _nilatomptr_; }
inline  AType NilAtom::atom_type() const { return NIL; }
inline  void  NilAtom::print( ostream& os ) const { os << "nil"; }
inline  void  NilAtom::sub_print( ostream& os ) const { os << ""; }
inline  UList NilAtom::eval( SymTable& ) { return nil; }

//--------------------------------------------------------------------------------
// RefAtom

inline  RefAtom::RefAtom(): Atom(), rc(1) {} 
inline  const  RefAtom *RefAtom::copy()    { ++rc; return this; }
inline  RefAtom        *RefAtom::destroy() 
{ 
    return ( --rc == 0 ) ? this : NULL; 
}

//--------------------------------------------------------------------------------
// ConsAtom

inline  ConsAtom::ConsAtom() {}
inline  ConsAtom::~ConsAtom() {}
inline  ConsAtom::ConsAtom( const ConsAtom& ca ): car(ca.car), cdr(ca.cdr) {}
inline  ConsAtom::ConsAtom( CRef A, CRef B ): car(A), cdr(B) {}
inline  UList  ConsAtom::head() const { return car; }
inline  UList  ConsAtom::tail() const { return cdr; }
inline  AType   ConsAtom::atom_type() const { return CONS; }

inline  ConsAtom  *ConsAtom::destroy()
{
#ifndef ALLOCATOR
    return  (--rc == 0) ? this : NULL;
#else
    if( --rc == 0 ) delete_cons(this);
    return NULL;
#endif
}

inline  const  ConsAtom *ConsAtom::clone() const
{
    return  new_cons( car.clone(), cdr.clone() );
}

inline  UList ConsAtom::eval( SymTable& env )
{
    return new_cons( car.eval(env), cdr.eval(env) ); 
}


//--------------------------------------------------------------------------------
// IntAtom

inline IntAtom::IntAtom( int n ): val(n) {}
inline IntAtom::~IntAtom() {}
inline  AType     IntAtom::atom_type() const { return INT; }
inline  int&      IntAtom::value() { return val; }
inline  const int IntAtom::value() const  { return val; }
inline  const IntAtom  *IntAtom::clone() const { return new_int(val); }
inline  IntAtom  *IntAtom::destroy()
{
#ifndef ALLOCATOR
    return  (--rc == 0) ? this : NULL;
#else
    if( --rc == 0 ) delete_int(this);
    return NULL;
#endif
}
inline  UList IntAtom::eval( SymTable& ) { return copy(); }

inline  void  IntAtom::print( ostream& os ) const { os << val; }
inline  void  IntAtom::sub_print( ostream& os ) const { os << ", " << val; }

// allow permissive extraction semantics
// UList     IntAtom::head() const  { return  const_cast<IntAtom*>(this)->copy(); } 

#endif 
//--------------------------------------------------------------------------------

//  Local Variables: ***
//  eval: (setq case-fold-search nil) ***
//  eval: (setq case-replace nil) ***
//  End: ***

