//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 14:55:38 sutner>
// 
// File:     util.cc
// Purpose:  implementation of atom oriented  UList --> UList  functions
// 
//--------------------------------------------------------------------------------

#include  "basic.h"
#include  "useratom.h"
#include  "util.h"
#include  "listutil.h"
#include  "sgroup.h"

//------------------------------------------------------------------------------
//  helpers

bool  ulist_to_transition( UList L, TransV& tr ) 
{
    UList  T, p, s, q;
    bool epsf = 1;
    tr.resize( 0 );
    while( L )
    {
	L >> T;
	T >> p >> s >> q;
	tr.push_back( Trans( ivalue(p)-1, ivalue(s)-1, ivalue(q)-1 ) );
	if( ivalue(s) == 0 ) epsf = 0;
    }
    return epsf;
}


void ulist_to_intarray( UList L, uvector& arr, int off ) 
{
    UList  x;
    arr.resize( 0 );
    while( L ) 
    {
	L >> x;
	arr.push_back( size_t(ivalue(x) + off) );
    }
}

UList  intarray_to_ulist( const uvector& arr, int off ) 
{
    UList L;
    for( int i=arr.size()-1; i>=0; i-- ) 
	L += (arr[i]+off);
    return L;
}

//----------------------------------------------------------------------

// ntriv==0: all components, ntriv==1: only non-trivial ones
// cmp:   characteristic vector of SCCs
// card:  cardinlities, 0 if trivial

UList  scc_to_ulist( const uvector& cmp, const uvector& card, int ntriv ) 
{
    UList  res;
    int  i, j, ncmp = card.size();
   
    vector<UList>  ptr(ncmp);
    for( i=cmp.size()-1; i>=0; i-- )
	ptr[cmp[i]] += (i+1);

    for( j = ncmp-1; j >= 0; j-- )
	if( !ntriv || card[j] )
	    res += ptr[j];

    return res;
}

//----------------------------------------------------------------------

void ulist_to_intlist( UList L, ulist& il, int off ) 
{
    UList  I;
    il.clear();
   
    while( L ) 
    {
	L >> I;
	il.push_back( size_t(ivalue(I) + off) );
    }
}

void ulist_to_ilist( UList L, ilist& il, int off ) 
{
    UList  I;
    il.clear();
   
    while( L ) 
    {
	L >> I;
	il.push_back( ivalue(I) + off );
    }
}

UList  intlist_to_ulist( const ilist& il, int off ) 
{
    UList L;
    ilist::const_iterator  itr;
    
    foreach_item(itr,il)
	L += ( *itr + off );
    return L;
}

//----------------------------------------------------------------------

void  ulist_to_intset( UList L, int msize, IntSet& set, int off ) 
{
    UList  p;
    set.initialize(msize);

    while( L )
    {
	L >> p;
	set.insert( ivalue(p) + off );
    }
}

UList  intset_to_ulist( const IntSet& set, int off )
{
    UList  L;
    IntSet::const_iterator  itr;

    foreach_item(itr,set)
	L += ( *itr+off );
    return L;
}
//----------------------------------------------------------------------

UList  bitvector_to_ulist( const BitVector& bv, int off )
{
    UList  L;

    for( int i=bv.size()-1; i>=0; i-- )
	L += ( bv.getbit(i)+off );
    return L;
}

UList  bitmatrix_to_ulist( const BitMatrix& bm, int off )
{
    UList  L;

    for( int i=bm.numRows()-1; i>=0; i-- )
	L += bitvector_to_ulist( *bm[i], off );
    return L;
}


//----------------------------------------------------------------------
// pre: size(st) == n or st==nil
// pre: reach subset (n)
// returns  st[reach] in proper order
// 
UList  order_select_ulist( UList st, IntSet& reach, int n )
{
    UList  res, h;
   
    vector<UList>  ptr(n);	// table of states
    int i = 0;

    if( st )
	while( st )
	{
	    st >> h;
	    ptr[i++] = h;
	}
    else			// or ints if st==nil
	for( i=0; i<n; i++ )
	    ptr[i] = new_int(i+1);

    IntSet::const_iterator  itr;
    foreach_item(itr,reach)
	res += ptr[*itr];

    return  reverse(res); 
}

//----------------------------------------------------------------------
//  part : (n) --> (m)  a partitioning map 
//  partition list L into classes according to part:  
//  L_i  is put into class  part[i] 
//

UList  partition_ulist( UList L, vector<int>& part, int n, int np )
{
    vector<UList>   Li(n);
    vector<UList>   LL(np);
    UList           res, h;   
    int             i = 0, j;   

    if( L )
	while( L )
	{
	    L >> h;
	    Li[i++] = h;
	}
    else
	for( i=0; i<n; i++ )
	    Li[i] = new_int(i+1);

    // distribute list elements
    for( i = 0; i < n; i++ )
    {
	j = part[i];
	LL[j] += Li[i];
    }

    for( i = 0; i < np; ++i ) 
	res += LL[i];

    return  reverse(res);
}

//--------------------------------------------------------------------------------

UList  size( UList L )
{  
    int  sz;
    
    switch( L.atom_type() )
    {
    case DFAA:  sz = dfavalue( L ).numStates(); break;
    case FAA:   sz = favalue( L ).numStates(); break;
    case GRA:   sz = grvalue( L ).numVertex(); break;
    case IARR:  sz = iavalue( L ).size(); break;
    case INT:   sz = ivalue( L ); break;
    case INTL:  sz = Ivalue( L ).num_bits(); break;
    case SAA:   sz = savalue( L ).numStates(); break;
    case SGA:   sz = sgvalue( L ).size(); break;
    case STR:   sz = strlen( svalue( L ) ); break;

    default:    sz = length_int( L );
    }
    
    return  new_int( sz );
}

//------------------------------------------------------------------------------
//  CA
//------------------------------------------------------------------------------

UList CAcompose( UList ca1, UList ca2 )
{
    UList  res = new CAAtom;
    cavalue( ca1 ).compose( cavalue(ca2), cavalue(res) );
    return res;
}


UList balanced( UList L, UList how )
{
    int  r = 1;
    if( how )  r = ivalue(how);
    return new_int( cavalue(L).balancedQ(r) );
}


UList CAanalyze( UList L )
{
    return new_int( cavalue( L ).analyze() );
}

// input:   { CA, pat, t, cyc }
// output:  orbit

UList CAorbit( UList L )
{
    UList     ca, patl, t, c;
    L >> ca >> patl >> t >> c;

    uvector  orb, pat;
    ulist_to_intarray( patl, pat ); 

    cavalue( ca ).orbit( pat, ivalue(t), ivalue(c), orb );
    return new IntArrAtom(orb);
}

// add conversion fa/dfa to sa 

UList  CAtoSA( UList ca )
{
    UList  res = new SAAtom();
    cavalue( ca ).toSA( savalue(res) );
    return res;
}

UList  CAtoWelch( UList L, UList how )
{
    int  norm = 1;
    if( how ) norm = ivalue( how );

    UList  sl = new SAAtom();
    UList  sr = new SAAtom();
    SA   sa; 

    cavalue( L ).toSA( sa );
    sa.welchAutomaton( savalue(sr), norm );
    sa.reverseAutomaton();
    sa.welchAutomaton( savalue(sl), norm );

    return  sl -= sr -= nil;
}

//----------------------------------------------------------------------
// Finite State Machines
//----------------------------------------------------------------------

static UList  delta_base( UList M, UList P, UList W, bool tr )
{
    UList  res;
    string  w = svalue(W);    
    ilist   p, pp;

    if( P.atom_type() == INT )
	p.push_back( ivalue(P)-1 );
    else
	ulist_to_ilist( P, p, -1 );

    if( tr ) res += P;
    
    if( M.atom_type() == SAA )
    {
	for( size_t i = 0; i < w.size(); ++i )
	{
	    savalue(M).delta( p, int(w[i] - 'a'), pp );
	    p = pp;
	    if( tr ) 
		res += reverse( intlist_to_ulist( p, 1 ) );
	}

	if( !tr ) res = intlist_to_ulist( p, 1 );
    }

    if( M.atom_type() == FAA )
    {
	for( size_t i = 0; i < w.size(); ++i )
	{
	    savalue(M).delta( p, int(w[i] - 'a'), pp );
	    p = pp;
	}
	res = intlist_to_ulist( p, 1 );
    }

    return reverse(res);
}


UList  delta( UList M, UList P, UList W )
{
    return  delta_base( M, P, W, 0 );
}

UList  Delta( UList M, UList P, UList W )
{
    return  delta_base( M, P, W, 1 );
}

	

UList  SAtoFA( UList sa )
{
    UList res = new FAAtom;
    savalue( sa ).toFA( favalue( res ) );
    return  res;
}

//----------------------------------------------------------------------

UList  FSMtoGraph( UList L )
{
    UList  res = new GraphAtom;
    
    if( L.atom_type() == SAA )
	savalue( L ).toGraph( grvalue(res) );

    if( L.atom_type() == FAA )
	favalue( L ).toGraph( grvalue(res) );
      
    if( L.atom_type() == DFAA )
	dfavalue( L ).toGraph( grvalue(res) );
      
    return res;
}

//----------------------------------------------------------------------

UList  reverseSA( UList fsm )
{
    UList  res = new SAAtom( savalue( fsm ) );
    savalue(res).reverseAutomaton();
    return res;
}

UList  reverseFA( UList fsm )
{
    UList  res = new FAAtom( favalue( fsm ) );
    favalue(res).reverseAutomaton();
    return res;
}

//----------------------------------------------------------------------
// A, B:  DFA
// how:   nil, int  or  (int,int)
UList  productDFA( UList A, UList B, UList how )
{
    int  norm = 1, unint = 1;
    switch( length_int( how ) )
    {
    case 1:   unint = ivalue( how ); break;

    case 2:   
	unint = ivalue( how.head() );
	norm  = ivalue( how.tail().head() );

    default:  break;
    }
    

    UList res = new DFAAtom;
    dfavalue( A ).productAutomaton( dfavalue( B ), dfavalue( res ), unint, norm );
    return  res;
}

// A, B:  DFA
// how:   nil or int
UList  quotientDFA( UList A, UList B, UList how )
{
    int  norm = 1;
    if( how ) norm = ivalue( how );

    UList res = new FAAtom;
    dfavalue( A ).quotientAutomaton( dfavalue( B ), favalue( res ), norm );
    return  res;
}

UList  squareSA( UList L )
{
    UList res = new SAAtom;
    savalue( L ).productAutomaton( savalue( res ) );
    return  res;
}

UList  productSA( UList A, UList B )
{
    UList res = new SAAtom;
    savalue( A ).productAutomaton( savalue( B ), savalue( res ) );
    return  res;
}

// pre:  L  list<list<int>>  in proper range

UList  kernelSA( UList S, UList L, UList how )
{
    int  norm = 1;
    if( how ) norm = ivalue( how );

    UList res = new SAAtom;
    savalue( S ).kernelAutomaton( savalue( res ), L, norm );
    return res;
}


//----------------------------------------------------------------------

UList  toDFA( UList L, UList how )
{
    int  norm = 1;
    if( how ) norm = ivalue(how);
   
    UList  res = new DFAAtom;

    if( L.atom_type() == FAA )
	favalue( L ).powerAutomaton(dfavalue( res ), norm );
    
    if( L.atom_type() == SAA )
    {
	FA      fa;
	savalue( L ).toFA(fa);
	fa.powerAutomaton(dfavalue( res ), norm );
    }

    return res;
}

//----------------------------------------------------------------------

UList  toDFASize( UList fsm )
{
    int   sz = 0;
   
    if( fsm.atom_type() == FAA )
	sz = favalue( fsm ).powerAutomatonSize();

    if( fsm.atom_type() == SAA )
    {
	FA    fa;
	savalue( fsm ).toFA(fa);
	sz = fa.powerAutomatonSize();
    }

    if( fsm.atom_type() == DFAA )
	sz = dfavalue( fsm ).numStates();

    return new_int(sz);
}

//----------------------------------------------------------------------

UList  FSMStates( UList fsm )
{
    if( fsm.atom_type() == SAA )
	return savalue( fsm ).states();

    if( fsm.atom_type() == FAA )
	return favalue( fsm ).states();

    if( fsm.atom_type() == DFAA )
	return dfavalue( fsm ).states();

    return nil;			// this cannot happen
}

//----------------------------------------------------------------------

UList  minimize( UList M, UList how )
{
    int    norm=1;
    if( how ) norm = ivalue( how );

    UList res = new DFAAtom;
    DFA  dfa;

    switch( M.atom_type() )
    {
    case SAA:
	savalue( M ).powerAutomaton( dfa, norm );
	dfa.minimalAutomaton( dfavalue( res ), norm );
	break;

    case FAA:
	favalue( M ).powerAutomaton( dfa, norm );
	dfa.minimalAutomaton( dfavalue( res ), norm );
	break;
	
    case DFAA:
	dfavalue( M ).minimalAutomaton( dfavalue( res ), norm );
	break;
	
    default: break;		// should not happen
    }

    return res;
}

//----------------------------------------------------------------------

UList  accessibleFSM( UList fsm, UList how )
{
    int    norm = 1;
    if( how ) norm = ivalue(how);
   
    if( fsm.atom_type() == DFAA )
    {
	UList res = new DFAAtom;      
	dfavalue( fsm ).accessiblePart( dfavalue( res ), norm );
	return res;
    }
    
    if( fsm.atom_type() == FAA )
    {
	UList res = new FAAtom;      
	favalue( fsm ).accessiblePart( favalue( res ), norm );
	return res;
    }
    
    return nil;			// this cannot happen
}

//----------------------------------------------------------------------

UList  DFAtoLanguage( UList L, UList R )
{
    return  dfavalue( L ).language( ivalue(R) );
}


//----------------------------------------------------------------------

UList  epsElimFA( UList L )
{
    UList  res = new FAAtom;      
    favalue( L ).epsilonElimination( favalue( res ) );
    return res;
}

UList  epsElimSA( UList L )
{
    UList  res = new SAAtom;      
    savalue( *L ).epsilonElimination( savalue( res ) );
    return res;
}


UList  Fischer( UList L )
{
    UList  sa = new SAAtom;
    III_Triple  t = savalue( L ).toFischer( savalue(sa) );
    UList  tt =  t.first() -= t.second() -= t.third() -= nil;
    return  tt -= sa -= nil;
}


UList  synchFischer( UList L )
{
    string w;
    int  p;
    p = synchronize_fischer( savalue(L), w );
//    return  new StringAtom( w.c_str() ) -= (p+1) -= nil;
    return  w -= (p+1) -= nil;
}


//----------------------------------------------------------------------
// semigroups
//----------------------------------------------------------------------


//  generators:  ( n, k, <flat int array> )

UList  toGeneratorsDFA( UList L )
{
    DFA   &dfa = dfavalue( L );
    return   dfa.numStates() -=  
	     dfa.numSymbols() -= 
	     intarray_to_ulist( dfa.transitions() );
}

//----------------------------------------------------------------------
// pre:  L = ( n, k, gen )  where gen is a flat list

class SGroupGen;

UList  generateSG( UList L, UList how )
{
    int norm = 1;
    if( how ) norm = ivalue( how );
    
    UList   ww, kk;
    L >> ww >> kk;

    uvector  gg(ivalue(ww) * ivalue(kk));
    ulist_to_intarray( L, gg ); 
    SGroupGen  gs( ivalue(ww), ivalue(kk), gg );

    UList  res = new SGAtom;
    switch( norm )
    {
    case 0:  gs.generate( sgvalue(res), SGroupGen::GENW ); break;
    case 1:  gs.generate( sgvalue(res), SGroupGen::GENE ); break;
    case 2: 
    default: gs.generate( sgvalue(res), SGroupGen::GEND ); break;
    }
    return res;
}


static UList equations_to_list( const SGroup& sg )
{
    list<Equation>::const_iterator itr; 
    UList res;
   
    foreach_item( itr, sg.equations() )
    {
	UList  pp;
	pp  = (*itr).first() -= (*itr).second()  -= nil;
	res = pp -= res;
    }
   
    return  reverse( res );
}

static UList witnesses_to_list( const SGroup& sg )
{
    vector<string>::const_iterator itr; 
    UList res;
   
    foreach_item( itr, sg.witnesses() )
	res = (*itr) -= res;
   
    return  reverse( res );
}


UList  equationListSG( UList L )
{
    return  equations_to_list( sgvalue(L) );
}

UList  witnessListSG( UList L )
{
    return  witnesses_to_list( sgvalue(L) );
}


static  UList  dclass_to_ulist( const DClass& dc )
{
    int  ww = dc.width();
    int  ks = dc.kSize();
    int  is = dc.iSize();
    int  hs = dc.hSize();

    UList res;

    for( int k=ks-1; k>=0; --k )
    {
	for( int i=is-1; i>=0; --i )
	{
	    for( int h=hs-1; h>=0; --h )
	    {
		for( int j=ww-1; j>=0; j-- )
		    res += dc.get( k,i,h,j ) + 1;
	    }
	}
    }
    
    res +=  ( ks -= is -= hs -= nil );
    return res;
}


UList  dClassListSG( UList S )
{
    assert( S.atom_type() == SGA );
    SGroup  &sg = sgvalue( S );
    
    UList res;
    for( int i = sg.dclasses().size() - 1; i >= 0; --i )
	res +=  dclass_to_ulist( sg.dclasses()[i] );
    return res;
}



//--------------------------------------------------------------------------------
// printing is side effect

UList  printElementsSG( UList L )
{
    sgvalue( L ).print_elements( cout );
    return nil;			
}


UList  printWitnessesSG( UList L )
{
    sgvalue( L ).print_witnesses( cout );
    return nil;			
}

UList  printEquationsSG( UList L )
{
    SGroup  &sg = sgvalue( L );
    
    if( sg.numEquations() )
	sg.print_equations( cout );
   
    return nil;
}

UList  printDClassesSG( UList L )
{
    SGroup  &sg = sgvalue( L );
    
    if( sg.numDClasses() )
	sg.print_dclasses( cout );
   
    return nil;
}

UList  aperiodicSG( UList L )
{
    SGroup  &sg = sgvalue( L );

    return new_int( sg.aperiodic() );
}

//----------------------------------------------------------------------
// graph algorithms 
//----------------------------------------------------------------------

UList  strongComponents( UList gr, UList ntr )
{
    int ntriv = 0;
    if( ntr ) ntriv = ivalue(ntr);
   
    uvector  part, card;
    grvalue( gr ).stronglyConnectedComponents( part, card );

    return scc_to_ulist( part, card, ntriv );
}

UList  longestPath( UList L )
{
    return new_int( grvalue(L).longestPath() );
}

//----------------------------------------------------------------------
// Integer related 
//----------------------------------------------------------------------

//Note:  LSD of n is in position 0 of A

void  integer_to_array( const Integer& n, uvector& A, const size_t base )
{
    Integer  nn(n);
    A.resize(0);
    while( nn > 0 )
    {
	A.push_back( (nn % int(base)).as_long() );
	nn /= int(base);
    }
}

void  array_to_integer( const uvector& A, Integer& n, const size_t base )
{
    n = 0;
    for( int i = A.size()-1; i >= 0; i-- )
	n = int(A[i]) + n * int(base);
}

//----------------------------------------------------------------------
