//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:34:59 sutner> 
//
//  :File:      zigzag.cc
//  :Purpose:   perform zigzag decomposition on a (2,2)-regular graph
//              generate permutation labeling
//              
//---------------------------------------------------------------------------

#include  <iostream>
#include  <fstream>
#include  <list>
#include  <vector>

#include  "stl-wrap.h"
#include  "getopt.h"
#include  "fsm.h"
#include  "basic.h"
#include  "urlist.h"
#include  "util.h"

//--------------------------------------------------------------------------------

const  char *zigzag_help_string = \
"h:     print this help message and quit\n"
"s:     switch label of first edge\n"
"S i:   switch label of edge number i\n"
"b i:   generate labeling according to number i\n"
"B bs:  generate labeling according to bit-sequence bs i\n"
"       read from/write to infile/outfile, defaults: cin, cout";


typedef  list<size_t>   ulist;	        // a zig-zag
typedef  vector<ulist>  adlist;	        // adjacency lists 

istream  *instrp;			// intput and output
ostream  *ostrp;    

// output formatting

char  *Format::lp = "(";
char  *Format::rp = ")";
char  *Format::cm = ", ";


//--------------------------------------------------------------------------------
// much too general for (2,2) regular graphs

inline  void   erase_one( ulist& L,  size_t s )
{
    ulist::iterator  it = find( L.begin(), L.end(), s );
    L.erase( it );
}

 size_t  next_node( size_t x, bool& dir, adlist& E, adlist& EE )
 {
     int z;

      if( dir )
     {
 	z = E[x].front();
 	E[x].pop_front();
 	erase_one( EE[z], x );
     }
     else
     {
 	z = EE[x].front();
 	EE[x].pop_front();
 	erase_one( E[z], x );
     }

     dir = !dir;
     return z;
 }

// relies on (2,2) regularity (zig-zags are even length)
size_t  zigzag_decomposition( size_t n, adlist& E, adlist& EE, list<ulist>& ZZ )
{
    bool dir = 1;
    ulist  L;
    size_t x, v = 0;

    while( v < n )
    {
	if( E[v].empty() )
	{
	    v++;
	    continue;
	}
	
	x = v;
	L.push_back(x);
	
	do
	{
	    x = next_node( x, dir, E, EE );
	    L.push_back(x);
	    x = next_node( x, dir, E, EE );
	    L.push_back(x);
	}
	while( x != v );

	ZZ.push_back( L );
	L.erase( L.begin(), L.end() );
    }
    
    return ZZ.size();
}


void  label_one( bool b, ulist& Z, SA* sa )
{
    bool  dir = 1;
    ulist::iterator it;
    int  z, x = Z.front();
    it = Z.begin();
    ++it;
    
    while( it != Z.end() )
    {
	z = *it++;
	if( dir )
	    sa->transitionsInsert( x, b, z );
	else
	    sa->transitionsInsert( z, b, x );
	b = !b;
	dir = !dir;
	x = z;
    }
}


// label all zig-zags, possibly switch the first label
// n:        number of nodes
// swi:      where to switch
// bits:     starting label for each zig-zag
// ZZ:       list of zig-zags

void  label_zigzags( int n, bool sw, int swi, ulist& bits, list<ulist>& ZZ, bool mt )
{
    ulist::iterator         it = bits.begin();
    list<ulist>::iterator  zit = ZZ.begin();

    if( !mt )
    {
	SA   sa( n, 2 );
	while( zit != ZZ.end() )
	    label_one( ( it != bits.end() ? *it++ : 0 ), *zit++, &sa );
	
	sa.sortTransitions( SA::STATE_ORDER );
	if( sw )
	   sa.transitions()[swi].s = !sa.transitions()[swi].s;

	write_atom_raw( *ostrp, sa, 1 );
    }
    else
    {
	FA   sa( n, 2 );
	while( zit != ZZ.end() )
	    label_one( ( it != bits.end() ? *it++ : 0 ), *zit++, &sa );
	sa.sortTransitions( SA::STATE_ORDER );
	if( sw )
	   sa.transitions()[swi].s = !sa.transitions()[swi].s;
	sa.initial().insert(0);
	sa.final().insert(0);
	write_atom_raw( *ostrp, sa, 1 );
    }

}


// reads from *instrp, performs no error checking 

size_t  read_graph( istream& is, adlist& E, adlist& EE )
{
    int n, x, y;
    string  s;
    is >> s;
    assert( s == "Graph" );
    
    is >> n;
    
    E.reserve( n );
    E.resize( n );
    EE.reserve( n );
    EE.resize( n );
    Fill( E, ulist() );
    Fill( EE, ulist() );

    for( int i = 0; i < n; ++i )
    {
	is  >> x >> y;
	E[i].push_back(x-1);
	E[i].push_back(y-1);
	EE[x-1].push_back(i);
	EE[y-1].push_back(i);
    }
    return n;
}


//--------------------------------------------------------------------------------


void  process_cmdline( int argc, char **argv, 
		       bool& sw, int& swi, bool& lb, ulist& bits, bool& mt )
{
    bool  numb = 0;
    bool  str  = 0;
    string       ss;
    int          ll(0);
    swi = 0;
    
    int cc;
    while( (cc=getopt(argc,argv,"hsfS:b:B:")) != EOF )
	switch( cc )
	{
	case 'h':
	    cerr << "Usage: " << argv[0];
	    cerr << " [-h] [-b i|B bs] [-s|S i] [-f] [infile] [outfile] " << endl;
	    cerr << zigzag_help_string << endl;
	    exit(0);
	    break;
	    
	case 's': 
	    sw = 1;
	    break;

	case 'f': 
	    mt = 1;
	    break;

	case 'S': 
	    sw = 1;
	    swi = atoi( optarg );
	    break;
	       
	case 'b': 
	    numb = lb = 1;
	    ll = atoi( optarg );
	    break;
	       
	case 'B':
	    str = lb = 1;
	    ss = optarg;
	    break;
	       
	default: break;
	}
	   
    if( optind < argc )
	instrp = new ifstream( argv[optind++]  );

    if( optind < argc )
	ostrp = new ofstream( argv[optind]  );

    if( !instrp )
	instrp = new istream(cin.rdbuf());

    if( !ostrp )
	ostrp = new ostream(cout.rdbuf());


    if( !instrp )
    {
	cerr << "cannot open input stream" << endl;
	exit(0);
    }

    if( !ostrp )
    {
	cerr << "cannot open output stream" << endl;
	exit(0);
    }

    if( numb )
	while( ll > 0 )
	{
	    bits.push_back( ll & 1U );
	    ll >>= 1;
	}
    
    if( str )
	for( int i = ss.size()-1; i >= 0; --i )
	    bits.push_back( ss[i] == '1' );
}

//--------------------------------------------------------------------------------
    
int  main( int argc, char **argv )
{
    ulist   bits;
    bool    label = 0;
    bool    sw = 0;
    bool    mt = 0;
    int     swi = 0;
    
    process_cmdline( argc, argv, sw, swi, label, bits, mt );

    vector<ulist>   E, EE;		// the adjacency lists

    size_t  n = read_graph( *instrp, E, EE );
    
    list<ulist>  ZZ;
    ulist::iterator  it;
    list<ulist>::iterator  iit;
    
    zigzag_decomposition( n, E, EE, ZZ );

    if( !label )		// zig-zag only
    {
	cout << ZZ.size() << endl;
	foreach_item( iit, ZZ )
	    {
		foreach_item( it, *iit )
		    cout << setw(4) << *it;
		cout << endl;
	    }
	exit(0);
    }
    
    label_zigzags( n, sw, swi, bits, ZZ, mt );

    delete ostrp;

    return EXIT_SUCCESS;
}




