(* ------------------------------------------------------------------------- *)
(* 
*  Time-stamp: <2004-01-15 09:29:14 sutner>
*
*  File:     experimental.m
*  Purpose:  new and unreliable macros 
*  Author:   K. Sutner, sutner@cs.cmu.edu
*
*)


(* ------------------------------------------------------------------------- *)
(* Combinatorics *)
(* ------------------------------------------------------------------------- *)


(* 
input:  W a list of binary words of the same length.

CheckSkeletal[W]  checks whether W is a Cartesian product.

TheSkeleton[W]   returns bitvector of free positions. 
*)

CheckSkeletal[W_List, K_ ] := 
Module[{tt,ch},
		alph = MakeAlphabet[K];
		If[ Length[W]==1, Return[True] ];
		If[ Length@Union[StringLength /@ W]>1, Return[False] ];
		tt = Thread[ ToCharacters /@ W ]; 
		ch = First[tt];
		WW = If[ Length@Union[tt[[#]]]==1, ch[[#]], alph ]& /@ 
					Range[Length[tt]] // Concatenate;
		WW == Sort[W]
];

TheSkeleton[ W_List, k_Integer ] := 
With[{tt = Thread[ ToCharacters /@ W ]}, 
		If[ Length@Union[tt[[#]]]==k, 1, 0 ]& /@ Range[Length[tt]] ];

(* ------------------------------------------------------------------------- *)

prot=Unprotect[Xor];
Xor[L_List,K_List] := Union[Complement[L,K],Complement[K,L]];
Protect[Evaluate[prot]];

(* ------------------------------------------------------------------------- *)

Options[TransientMod] = {Index -> 0};

TransientMod[ x_Integer, t_Integer?NonNegative, p_Integer?Positive, opts___?OptionQ ] :=
With[ { tt = t + (Index /. {opts} /. Options[TransientMod]) },
      	If[x < tt, x, Mod[x - tt, p] + tt]
];

TransientMod[ x_Integer, {t_Integer?NonNegative, p_Integer?Positive}, opts___?OptionQ ] :=
TransientMod[ x, t, p, opts ];

NextSubset[ L_List, n_Integer ] := 
Module[ {i},
        i = -1;
        While [ L[[i]] == n + i + 1, i-- ]; 
        Join[  Drop[ L, i ], Range[-i] + L[[i]] ]
] /;  Length[L] <= n && L =!= Range[ n - Length[L] + 1, n ];

NextSubset[ L_List, n_Integer ] :=  Range[Length[L]]   /;  Length[L] <= n ;

NextPermutation[ LL_List ] := 
Module[ {a,b,L,n,i,j,jj},

	If[ LL==Reverse@Sort[LL], Return[Reverse[LL]] ];

	n = Length[LL];
	i = -2;
	While[ LL[[i]] >= LL[[i+1]], i-- ];
	
	a = LL[[i]];
	j = jj = i+1;
	b = LL[[jj]];
	While[ jj <= -1, 
		If[ b >= LL[[jj]] > a, b = LL[[jj]]; j = jj ];
		jj++
	];
	
	(* flip a and b, sort second part of list *)
	L = ReplacePart[ ReplacePart[LL,a,j],b,i];
	Join[ Drop[L,i+1], Reverse[Take[L,i+1]] ]
];

NextPermutation[ LL_List, n_Integer ] := NestList[ NextPermutation, LL, n ];

WordSuccessor[ ww_?WordQ ] := 
With[ {alph = Alphabet[ww]},
	ToWord[ NextPermutation[ToIndex[ww,alph]], alph ] 
];

(* ------------------------------------------------------------------------- *)

Options[WordOrbit]={Full->True};

WordOrbit[ww_?WordQ,opts___?OptionQ] := 
With[ {alph = Alphabet[ww]},
	If[ Full /. {opts} /. Options[WordOrbit],
		ToWord[#,alph]& /@  
			NextPermutation[ ToIndex[ww,alph], 
					Multinomial@@ParikhVector[ww,alph]-1 ],
	(* else *) 
		Multinomial@@ParikhVector[ww,alph]
	]
];

WordSubwordQ[Eps | "Eps" | "" | Global`\[Epsilon], _?WordQ] := True;
WordSubwordQ[x_?WordQ, y_?WordQ] := 
    Module[ {alph},
        	alph = Union[Alphabet[x], Alphabet[y]];
        	LongestCommonSubsequence[ToIndex[x, alph], ToIndex[y, alph], 
            Full -> False] == WordLength[x]
        ] /; WordLength[x] <= WordLength[y];
WordSubwordQ[x_?WordQ, y_?WordQ] :=  False  /; WordLength[x] > WordLength[y];


(* ------------------------------------------------------------------------- *)
(* Polya Counting *)

ClearAll[CycleIndexPolynomial, SubstitutePolynomial]

CycleIndexPolynomial[t_T] := 
Module[{n = Length[t]}, 
    cd = Length /@ CycleDecompositionT[t, Full -> True]; 
    cnt = (Count[cd, #1] & ) /@ Range[n]; 
    Times @@ (Array[Subscript[x, #1] & , {n}]^cnt)
]; 

CycleIndexPolynomial[S_SG] := Plus @@ CycleIndexPolynomial /@ ToList[S]/Length[S]; 

Options[SubstitutePolynomial] = {Full -> True}; 
SubstitutePolynomial[pol_, n_Integer, m_Integer | m_Symbol, (opts___)?OptionQ] := 
If[ Full /. {opts} /. Options[SubstitutePolynomial], 
    pol /. Table[Subscript[x, i] -> m, {i, n}], 
    pol /. Table[Subscript[x, i] -> Plus @@ Array[Subscript[c, #1]^i & , {m}], {i, n}]
]; 


(* ------------------------------------------------------------------------- *)


ListLessEqual[ L_List, K_List ] := And@@Thread[ L <= K ]   /; Length[L]==Length[K];
ListLess[ L_List, K_List ] := ListLessEqual[ L, K ] && L =!=K   /; Length[L]==Length[K];

MatrixLessEqual[ L_?MatrixQ, K_?MatrixQ ] := 
(L <= K //. LessEqual -> ListLessEqual)   /;  Dimensions[L] == Dimensions[K];
MatrixLess[ L_?MatrixQ, K_?MatrixQ ] := 
(L <= K //. Less -> ListLess)   /;  Dimensions[L] == Dimensions[K];


(* ------------------------------------------------------------------------- *)
(* Graphics, requires Colors` *)
(* ------------------------------------------------------------------------- *)

DisplayColors[]:=
Module[ {vsize,hsize,splatch},
	vsize = 25;
	hsize = 150;
	splatch[cc_,{x_,y_}] := 
	{ ToExpression[cc], Rectangle[{x hsize,y vsize},{(x+.25)hsize,(y+1)vsize}], 
 	  Black, Text[ cc,{x hsize+40,y vsize+vsize/2},{-1,0}] };
          
	  Show[  Graphics@MapIndexed[ splatch, 
		          Reverse /@ Partition[ AllColors,50,50,1,{}],{2} ],
	PlotRange->{{hsize,6 hsize},{vsize,50*vsize}}, AspectRatio->Automatic];
];

(* plot ECA *)


LineClosed[L_List] := Line[Append[L, First[L]]];

plotecabbox =  { Black, Thickness[.01], 
      LineClosed[{{0, 1}, {3, 1}, {3, 2}, {0, 2}}],
      Line[{{1, 1}, {1, 2}}], Line[ {{2, 1}, {2, 2}}], Red, 
      LineClosed[ {{1, 0}, {2, 0}, {2, 1}, {1, 1}}] };

plotecamakebox[zz_] := 
    With[{z = 4zz}, {Thickness[.002], Black, 
        LineClosed[ (# + {z, 0}) & /@ {{0, 1}, {3, 1}, {3, 2}, {0, 2}}],
        Line[(# + {z, 0}) & /@ {{1, 1}, {1, 2}}], 
        Line[(# + {z, 0}) & /@ {{2, 1}, {2, 2}}], Red, 
        LineClosed[(# + {z, 0}) & /@ {{1, 0}, {2, 0}, {2, 1}, {1, 1}}] } ];

plotecamakefields[ v_, zz_] := 
    Module[ {cols, a, b, c, z},
      z = 4zz;
      cols = {RGBColor[0.8, 1., 1.], RGBColor[0., 0., 1.], GrayLevel[0.8]};
      {a, b, c} = IntegerDigits[zz, 2, 3] + 1;
      { cols[[a]], Rectangle[ {z, 1},        {z + 1, 2}], 
         cols[[b]], Rectangle[  {z + 1, 1}, {z + 2, 2}], 
         cols[[c]], Rectangle[  {z + 2, 1}, {z + 3, 2}],
         cols[[v + 1]], Rectangle[  {z + 1, 0}, {z + 2, 1}]  } ];

PlotECA[ L_List ] := 
    Show[ Graphics[ Flatten[  
          Table[  { plotecamakefields[L[[i + 1]], i], plotecamakebox[i] }, {i,0,7} ]]],
      AspectRatio -> .07];
PlotECA[ r_Integer ] := PlotECA[ Reverse@IntegerDigits[r, 2, 8]];


(* ------------------------------------------------------------------------- *)
(* Algebra *)
(* ------------------------------------------------------------------------- *)

ClearAll[PadicOrder, PadicValuation];

PadicOrder[0, _] = Infinity;
PadicOrder[L_List, p_Integer?(#>1&)] := PadicOrder[#, p] & /@ L;
PadicOrder[x_Integer, p_Integer?(#>1&)] := 
    Module[{e = 0, xx = x},
      	While[ Mod[xx, p] == 0, e++; xx /= p ];
      	e
];
PadicOrder[x_Rational, p_Integer?(#>1&)] := 
    	PadicOrder[Numerator[x], p] - PadicOrder[Denominator[x], p];

PadicValuation[0, _] = 0;
PadicValuation[x_Integer | x_Rational, p_Integer?(#>1&)] := p^(-PadicOrder[x, p]);

(* ------------------------------------------------------------------------- *)
(* actions for lists and words *)

ClearAll[ListMultiply];
ListMultiply[ K_List, 0 ] :=  K;
ListMultiply[ K_List, 1 ] :=  1 - K;
ListMultiply[ K_List, L_List ] := Join@@( ListMultiply[K,#]& /@ L );

(* u acts on w, result has length  |w|*|u|
   u == 0  neutral, u == 1 flips all bits  (k==2) 
*)
wordmult[k_Integer][ L_List, s_Integer ] := Mod[ L+s, k ];
WordAct[k_Integer][ w_?WordQ, u_?WordQ ] := 
With[ {ind = wordmult[k][ ToExpression /@ ToCharacters[w], # ]& /@ 
				ToExpression /@ ToCharacters[u]},
		StringJoin@@( ToString /@ Flatten[ind] )
];

ClearAll[ActionT];
ActionT[LL_List, T[x__]] := LL[[{x}]];
ActionT[LL_List, {tt___T}] := (ActionT[LL, #1] &) /@ {tt};
ActionT[LL_List, S_SG] := ActionT[LL, ToList[S]];


<< LinearAlgebra`MatrixManipulation`;
KroneckerProduct[ AA_?MatrixQ, BB_?MatrixQ ] :=
With[ {dims = Dimensions[AA]},
    BlockMatrix[  AA Table[  BB, {dims[[1]]}, {dims[[2]]} ] ]
];


KroneckerProduct[ AA_?MatrixQ, BB_?MatrixQ ] := 
    BlockMatrix[  AA Array[  BB&,  Dimensions[AA] ] ];


KroneckerProduct[ AA_?MatrixQ, BB_?MatrixQ, xx___?MatrixQ ] := 
    KroneckerProduct[ KroneckerProduct[AA,BB], xx ] ; 

Attributes[WitnessFunction]={HoldFirst};
WitnessFunction[ gg_, S_SG, W_List ] := 
		 WitnessFunction[ gg, ToList[S], W ] /; Length[S]==Length[W];
WitnessFunction[ gg_, S_List, W_List ] := 
(
	AssignFunction[ gg, S, W, Opts->Listable ];
	gg[DC[xx_]] := DC[gg[xx]]
);


(* ------------------------------------------------------------------------- *)
(* FSM *)
(* ------------------------------------------------------------------------- *)


AddState[ m_FA | m_SA ] := 
Module[ {n, mm},
    	n = Size[m];
    	mm = ReplacePart[ m, n + 1, 1 ];
    	If[  MinusEins[mm] === {}, Return[mm]];
    	ReplacePart[ mm, {}, -1 ]
];

AddState[ m_DFA ] := AddState[ToFA[m]];
	

CommutativeQFA[ m_?FSMQ ] :=
Module[{mm,p,sa},
	mm = ToNFA[m];
	If[Size[mm]==1,Return[True]];
	p = TheSinkDFA[mm];
	If[p==0,Return[False]];
	sa = ToSA@DeleteStateFA[ mm, p];
	IsomorphicQDFA[ mm, MinimizeFA[sa]]
];

LassoDFA[t_Integer, p_Integer, F_List] := 
    With[ {n = t + p},
      DFA[ n, 1, {Join[ Range[2, n], {t + 1}]}, 1, F ] ];

LassoDFA[t_Integer, p_Integer, q_Integer] := LassoDFA[t, p, {q}];

AllLassoDFA[  t_Integer, p_Integer ] := LassoDFA[t, p, #] & /@ PowerSet[Range[t+p]];

AllLassoDFA[  n_Integer ] := FlattenOne[Table[AllLassoDFA[i,n-i],{i,0,n-1}]] ;

PermutationDFA[ perms_List, K_:Automata`automata`$alphabet ] := 
With[ { n = Length@First[perms] }, 
	DFA[ n, Automata`automata`Private`ShortAlphabetSize[K], perms, 1, {1}, 
	Automata`automata`Private`ShortAlphabet[K], {} ]
];

ToSAGraph[ g_Graph, lab_List, K_:Automata`automata`$alphabet] := 
With[ {ed = Edges@ToEdgeList[g]}, 
     SA[ NumVertices[g], Automata`automata`Private`ShortAlphabetSize[K], 
     	Thread[{First/@ed, lab, Last/@ed}], {}, {} ]
];

Options[ExtFischer]={Full->False};
ExtFischer[s_SA,opts___?OptionQ] := 
If[ Full /. {opts} /. Options[ExtFischer], 
	Automata`extaut`SendProgram["fischer[", s, "]" ], 
	Automata`extaut`SendProgram["*fischer[", s, "]" ] 
];


(* expects LL to be a list of words of uniform length n >> d *)
ConstructSA[ LL_List, d_Integer, k_ ] := 
	Module[ {quot,trans,h,K,L,L0,U,Q,nn},
	K = MakeAlphabet[k];
	L = ToString /@ LL;
	L0 = Union[ StringDrop[#,-1]& /@ L ];
	U = Union[ StringTake[#,d]& /@ L ];
	quot[x_] := quot[x] = 
		Union[ StringDrop[#,d]& /@ Select[L0,StringTake[#,d]===x&] ];
	Q = Union[quot /@ U];
	nn = Length[Q];
	AssignFunction[ h, Q, Range[nn] ];
	h[_] = 0;
	
	trans[x_,i_] := trans[x,i] = 
		With[ {xx = StringJoin[x,K[[i]]]},
		 {h[quot[x]],i, h[Union[ StringDrop[#,d+1]& /@
		 	Select[L,StringTake[#,d+1]===xx&] ]]} ];
	tr = Union@Select[ Apply[ trans, CartesianProduct[U,Range[k]], {1} ], 
			    Last[#]>0& ];
	SA[ nn, k, tr ]
];


(* ------------------------------------------------------------------------- *)
(* Convert a list of paths *)

(*  select ONE word associated with path *)
PathToWord[ M_?FSMQ, {LL___List} ] := 
With[ {tr = FullTransitions@ToFA[M]},
	Concatenate[ Function[xx, First@Cases[ tr, {First[xx],aa_,Last[xx]}->aa ]] /@ 
			Partition[#,2,1] ]& /@ {LL}
];

PathToWord[ M_?FSMQ, {_} ] := "Eps";
PathToWord[ M_?FSMQ, LL_List ] := 
With[ {tr = FullTransitions@ToFA[M]},
	Concatenate@@
	(First@Cases[ tr, {First[#],aa_,Last[#]}->aa ]& /@ Partition[LL,2,1])
];

(* 
fragment for all words associated with a path:

toword[{_}] := {"Eps"};
toword[{p_,q_}] := Cases[ tr, {p,aa_,q}->aa ];
toword[LL_List] := 
	Concatenate[ Function[xx, Cases[ tr, {First[xx],aa_,Last[xx]}->aa ]] /@ 
			Partition[LL,2,1] ];
*)


(* ------------------------------------------------------------------------- *)
(* Transducers *)
(* ------------------------------------------------------------------------- *)



cp2num[n_][{i_,j_}] := (i-1)n + j;
num2cp[n_][ii_] :=  {  Quotient[ii-1,n]+1, Mod[ii-1,n]+1 };


Options[ApplyTD]={Normalize->0};

ApplyTD[ M_?FSMQ, tr_TD, opts___?OptionQ ] := 
Module[{k1,k2,n1,n2,tm,tt,combine,sel1,sel2,t0,ttt,res},
	k1 = AlphabetSize[M];
	k2 = Drei[tr];
	n1 = Size[M];
	n2 = First[tr];
	tm = Transitions[ToSA[M]];
	tt = tr[[4]];
	ini = tr[[5]];
	fin = tr[[6]];
	combine[{{p_,pp_},{q_,b_,qq_}}] := {cp2num[n2][{p,q}],b,cp2num[n2][{pp,qq}]};
	sel1[a_] := Cases[tm,{p_,a,pp_}->{p,pp}];
	sel2[a_] := Cases[tt,{q_,a,b_,qq_}->{q,b,qq}];
	t0 = combine /@ CartesianProduct[ Union[ sel1[0], 
	     Pairs[Range[n1],Range[n1]]],sel2[0]];
	ttt = {};
	AppendTo[ttt,combine/@CartesianProduct[sel1[#],sel2[#]]]& /@ Range[k1];

	res = SA[ n1 n2, k2, Join[ t0, FlattenOne[ttt] ] ];

	If[ Head[M]=!=SA,
	    res = SetInitialFinalFA[ res, 
		    cp2num[n2] /@ CartesianProduct[Initial[M],ini], 
		    cp2num[n2] /@ CartesianProduct[Final[M],fin] ]
	];
	Switch[ Normalize /. {opts} /. Options[ApplyTD],
	    0,	res,
	    1,  ReplacePart[ res, CartesianProduct[n1,n2], -1 ],
	    2,  ReplacePart[ res, CartesianProduct[States[M],Range[n2]], -1 ]
	]
];


ToTransducerCA[ C_CA ] := 
With[ {k = AlphabetSize[C],w=WidthCA[C],
kk = Automata`automata`Private`ShortAlphabetSize[C]},

	TD[ k^w, kk, kk, 
	Thread@Insert[Thread@Transitions@ToSA[C],Flatten@Table[i,{k^(w-1)},{i,k}],2],
	Range[k^w], Range[k^w] ] 
];

splitTrans[K1_, K2_][{p_,u_,v_,pp_}, n_] := 
Module[ {uu,vv,lu,lv,ll},
	uu = ToIndex[u,K1];
	vv = ToIndex[v,K2];
	lu = Length[uu];
	lv = Length[vv];
	ll = Max[lu,lv,1];
	uu = Join[uu,Table[0,{ll-lu}]];
	vv = Join[vv,Table[0,{ll-lv}]];
	{ Pairs[   Join[{p}, Range[n+1,n+ll-1]], uu, vv, Join[ Range[n+1,n+ll-1],{pp}] ],
	  n + ll - 1 }
];

SplitTransformations[ M_TD ] :=
Module[ {ttt,tt,t,nn,K1,K2},
	nn = First[M];
	K1 = Zwei[M];
	K2 = Drei[M];
	tt = M[[4]];
	ttt = {};
	Do[
		{t,nn} = splitTrans[K1,K2][ tt[[i]], nn ];
		ttt = Join[ ttt, t ],
		{i,Length[tt]}
	];
	ReplaceParts[ M, nn, 1, ttt, 4 ]
];


(* ------------------------------------------------------------------------- *)
(* Graphs *)
(* ------------------------------------------------------------------------- *)

DistanceGraph[ G_Graph, v_Integer ] :=
Module[ {GG, dist, Q, A, x},
   	GG = ToAdjacencyList[G];
   	A = Edges[GG];
   	dist[v] = 0;
   	dist[_] := Infinity;
   	Q = {v};
   
   	While[ Q =!= {},
     		{x, Q} = {First[Q], Rest[Q]};
     		
		If[ dist[#] == Infinity, 
		    dist[#] = dist[x] + 1; AppendTo[Q, #]] & /@ A[[x]];
     	];
   
   	dist /@ Range[NumVertices[GG]]
]  /;  1 <= v <= NumVertices[G];




GraphToSA[ G_Graph, L_List, k_Integer ] := 
With[ {ed = Edges@ToEdgeList[G]},
      SA[ NumVertices[G], k, Thread[{ First/@ed,L,Last/@ed}], {}, MinusEins[G]]]

GraphToSA[ G_Graph,ff_, k_Integer ] := 
With[ {ed = Edges@ToEdgeList[G]},
      SA[ NumVertices[G], k, {#[[1]],ff[#[[1]],#[[2]]],#[[2]]}& /@ ed, {}, MinusEins[G]]]


ZigzagDecompositionGraph::usage = "ZigzagDecompositionGraph[G] computes the zigzag decomposition of a transitive (2,2)-regular graph G.";

ZigzagDecompositionGraph[G_Graph] :=
Module[{adj, adjr, zig, zag, makezz, ed, e1, e2, Z, Z1},
      adj = Edges[ToAdjacencyList[G]];
      adjr = Edges[ToAdjacencyList[ReverseGraph[G]]];

      zig[{p_, q_}] :=
        If[adj[[p]] === {q, q}, 
		    {p, q}, 
		    {p, First[Complement[adj[[p]], {q}]]}];
      zag[{p_, q_}] :=
        If[adjr[[q]] === {p, p}, 
		     {p, q}, 
		     {First[Complement[adjr[[q]], {p}]], q}];
      makezz[{p_, q_}] :=
      (
          	Z1 = {{p, q}, zag[{p, q}]};
          	While[Length[Z1] == 2 || Take[Z1, 2] =!= Take[Z1, -2],
            		e1 = zig[Last[Z1]];
            		e2 = zag[e1];
            		Z1 = Join[Z1, {e1}, {e2}];
            	];
          	Drop[Z1, -2]
      );

      ed = Edges[ToEdgeList[G]];
      Reap[
          While[ ed =!= {},
            	Z = makezz[First[ed]];
            	Sow[Z];
            	ed = Complement[ed, Z];
      ]][[2, 1]]
];


(* ------------------------------------------------------------------------- *)
(* Regular Expressions *)
(* ------------------------------------------------------------------------- *)

REPower[_,0] := Eps;
REPower[e_,n_Integer?Positive] := RET@@Table[e,{n}];

PrintRE[re_] := ToRegExp[re] //. {RES[x_]->x^"*",Empty->\[EmptySet]};

(* ------------------------------------------------------------------------- *)
(* Miscellany *)
(* ------------------------------------------------------------------------- *)

ToGraphT[ xx_T ] := 
With[ {n = Length[xx]},
		Graph[ n, n, "EList", Pairs[Range[n],ToList[xx]],{}] ];



RandomInteger[m_Integer:1] := Random[Integer,{0,m}];
RandomInteger[n_Integer:0,m_Integer:1] := Random[Integer,{n,m}];

(* ------------------------------------------------------------------------- *)

Options[MarkPeriod]={Marker->{0->2,1->5}};
MarkPeriod[L_List,opts___?OptionQ]:=
Module[{qq,pp,mm,K},
  	qq = Flatten/@Cases[Position[L,#]&/@Union[L],{_,__}];
  	If[ qq==={}, Return[L] ];
  	pp = First@Sort[First/@qq];
  	K = L[[pp]] /. (Marker /. {opts} /. Options[MarkPeriod]);
  	L /. L[[pp]]->K
];


(* ------------------------------------------------------------------------- *)

writeStringList[ os_, L_List ] :=
   ( WriteString[os,#,"\t"]& /@ Drop[ L,-1]; WriteString[os,Last[L],"\n"]);


(* ------------------------------------------------------------------------- *)
(* new formatting *)


$Post = ReplaceRepeated[#,{""->\[Epsilon],Eps->\[Epsilon],"Eps"->\[Epsilon]}]&;

CirclePlus = Concatenate;
EnumForm0 = Curry[EnumForm][Index -> 0];

MakeAbbrevs::msgon = "Abbreviations turned on, use MakeAbbrevs[False] to turn off.";
MakeAbbrevs[b_:True] := 
(
    LFA = LanguageFA;
    TF  = TableForm;
    TF2 = TableForm[#,TableDepth->2]&;
    TF1 = TableForm[#,TableSpacing->{1,1}]&;
    CF  = ColumnForm;
    MF  = MatrixForm;
    CP  = CartesianProduct;
    SCC = StronglyConnectedComponents;
    REFA = RegularExpressionToFA;
    CodeterministicQFA = DeterministicQFA[#,Direction->Backward]&;

    Message[MakeAbbrevs::msgon];
);

MakeAbbrevs[False|None] := 
Clear[LFA, TF, TF1, TF2, CF, MF, CP, SCC, REFA, CodeterministicQFA];


(* ------------------------------------------------------------------------- *)
(* Conversion *)
(* ------------------------------------------------------------------------- *)

TeXStates[ M_?FSMQ ] := 
If[ IntegerQ@First[M], 
	Size[M] // ToString,
	States[M] // TeXForm // ToString ];

TeXFormFA[ M_FA ] := 
StringJoin[ 
	"\\langle [", TeXStates[M], "], ",
	Alphabet[M] // TeXForm // ToString, ", ",
	"\\tau", ", ",
	Initial[M] // TeXForm // ToString, ", ",
	 Final[M] // TeXForm // ToString, "\\rangle \n\n", 
	 "$$\n", 
	 StringReplace[ Trans[M] // Thread // MF // TeXForm // ToString, 
	 	{"Eps"->"\\eps", "\\cr"->"\\cr\n"} ],
	 "\n$$\n" ];
	
TeXFormFA[ M_DFA ] := 
StringJoin[ 
	"\\langle [", TeXStates[M], "], ",
	Alphabet[M] // TeXForm // ToString, ", ",
	"\\delta", ", ",
	Initial1[M] // TeXForm // ToString, ", ",
	 Final[M] // TeXForm // ToString, "\\rangle \n\n", 
	 "$$\n", 
	 StringReplace[ Trans[M] // MF // TeXForm // ToString, 
	 	{"Eps"->"\\eps", "\\cr"->"\\cr\n"} ],
	 "\n$$\n" 
];
	
ClearAll[toVaucanson]
toVaucanson[M_?FSMQ, fl_String, opts___?OptionQ ] :=
  
  Module[{n, k, alph, vlab, stlab, tr, st, ostrm, trf, writeStates, 
      writeTrans,statecoord},
    
    statecoord[p_] := 
      If[ Full /. {opts} /. Options[ToVaucanson], 
        StringJoin["(", "0", ",", ToString[p], ")"],
        StringJoin["(,)"]
        ];
    
    writeStates[p_] := 
      WriteString[ostrm, "\\State[", stlab[[p]], "]{", statecoord[p],
        "}{", vlab[[p]], "}\n"];
    
    writeTrans[{p_, p_}] := 
      With[{labs = trf[{p, p}]},
        WriteString[ostrm, "\\LoopN{", vlab[[p]], "}"];
        writeStringList[ostrm, alph[[labs]] ];
        ];

    writeTrans[{p_, q_}] := 
      With[{labs = trf[{p, q}]},
        WriteString[ostrm, "\\EdgeL{", vlab[[p]], "}{", vlab[[q]], "}"];
        writeStringList[ostrm, alph[[labs]] ];
        ];
    
    writeStringList[os_, L_List, sep_String:","] :=
      (
        WriteString[os, "{"];
        WriteString[os, #, sep] & /@ Most[L];
        WriteString[os, Last[L]];
        WriteString[os, "}\n"]
        );
    
    	n = Size[M];
    	{k, alph} = Alphabet[M, Full -> True];
    	vlab = FromCharacterCode /@ Range[65, 64 + n];
    	stlab = StringReplace[ToString[#], " " -> ""] & /@ States[M] ;
    	tr = Transitions[ToFA[M]];
    	trf[{_, _}] := {};
    	Scan[  AppendTo[trf[{#[[1]], #[[3]]}], #[[2]]] &, tr ];
    	st = Union[Delete[#, {2}] & /@ tr];
    	ostrm = OpenWrite[fl <> ".vcs"];
    	writeStates /@ Range[n];
    	WriteString[ostrm, "%\n"];
    	writeTrans /@ st;
    	Close[ostrm]
    ];
		

Null

(* ------------------------------------------------------------------------- *)
(* eof *)
